from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget
from superqt import QLabeledDoubleSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr
from napari.layers.base.base import Layer
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtGammaSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current gamma
    attribute value and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Layer
        An instance of a napari layer.

    Attributes
    ----------
    gamma_slider : superqt.QLabeledDoubleSlider
        Gamma adjustment slider widget.
    gamma_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the gamma chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Layer) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        sld = QLabeledDoubleSlider(Qt.Orientation.Horizontal, parent)
        sld.setMinimum(0.2)
        sld.setMaximum(2)
        sld.setSingleStep(0.02)
        sld.setValue(self._layer.gamma)
        connect_setattr(sld.valueChanged, self._layer, 'gamma')
        self._callbacks.append(
            attr_to_settr(self._layer, 'gamma', sld, 'setValue')
        )
        self.gamma_slider = sld

        self.gamma_slider_label = QtWrappedLabel(trans._('gamma:'))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.gamma_slider_label, self.gamma_slider)]



"""
Annotate segmentation with text
===============================

Perform a segmentation and annotate the results with
bounding boxes and text.
This example is fully explained in the following tutorial:
https://napari.org/stable/tutorials/segmentation/annotate_segmentation.html

.. tags:: analysis
"""
import numpy as np
from skimage import data
from skimage.filters import threshold_otsu
from skimage.measure import label, regionprops_table
from skimage.morphology import closing, remove_small_objects, square
from skimage.segmentation import clear_border

import napari


def segment(image):
    """Segment an image using an intensity threshold determined via
    Otsu's method.

    Parameters
    ----------
    image : np.ndarray
        The image to be segmented

    Returns
    -------
    label_image : np.ndarray
        The resulting image where each detected object labeled with a unique integer.
    """
    # apply threshold
    thresh = threshold_otsu(image)
    bw = closing(image > thresh, square(4))

    # remove artifacts connected to image border
    cleared = remove_small_objects(clear_border(bw), 20)

    # label image regions
    label_image = label(cleared)

    return label_image


def make_bbox(bbox_extents):
    """Get the coordinates of the corners of a
    bounding box from the extents

    Parameters
    ----------
    bbox_extents : list (4xN)
        List of the extents of the bounding boxes for each of the N regions.
        Should be ordered: [min_row, min_column, max_row, max_column]

    Returns
    -------
    bbox_rect : np.ndarray
        The corners of the bounding box. Can be input directly into a
        napari Shapes layer.
    """
    minr = bbox_extents[0]
    minc = bbox_extents[1]
    maxr = bbox_extents[2]
    maxc = bbox_extents[3]

    bbox_rect = np.array(
        [[minr, minc], [maxr, minc], [maxr, maxc], [minr, maxc]]
    )
    bbox_rect = np.moveaxis(bbox_rect, 2, 0)

    return bbox_rect


def circularity(perimeter, area):
    """Calculate the circularity of the region

    Parameters
    ----------
    perimeter : float
        the perimeter of the region
    area : float
        the area of the region

    Returns
    -------
    circularity : float
        The circularity of the region as defined by 4*pi*area / perimeter^2
    """
    circularity = 4 * np.pi * area / (perimeter ** 2)

    return circularity


# load the image and segment it
image = data.coins()[50:-50, 50:-50]
label_image = segment(image)

# create the features dictionary
features = regionprops_table(
    label_image, properties=('label', 'bbox', 'perimeter', 'area')
)
features['circularity'] = circularity(
    features['perimeter'], features['area']
)

# create the bounding box rectangles
bbox_rects = make_bbox([features[f'bbox-{i}'] for i in range(4)])

# specify the display parameters for the text
text_parameters = {
    'string': 'label: {label}\ncirc: {circularity:.2f}',
    'size': 12,
    'color': 'green',
    'anchor': 'upper_left',
    'translation': [-3, 0],
}

# initialise viewer with coins image
viewer = napari.Viewer()
layer = viewer.add_image(image, name='coins', rgb=False)

# add the labels
label_layer = viewer.add_labels(label_image, name='segmentation')

shapes_layer = viewer.add_shapes(
    bbox_rects,
    face_color='transparent',
    edge_color='green',
    features=features,
    text=text_parameters,
    name='bounding box',
)

if __name__ == '__main__':
    napari.run()



"""Example script for starting napari with a custom colormap

This is an example script that could be provided in napari setting
to be executed on napari application startup.

This script adds a custom colormap named 'cyan_t' to the napari
"""

from napari.utils.colormaps import Colormap, ensure_colormap

custom_cyan_colormap = Colormap(colors=ensure_colormap('cyan').colors, name='cyan_t', low_color=[0, 0, 0, 0])
ensure_colormap(custom_cyan_colormap)



import argparse
import importlib
from typing import NamedTuple

from .utils import run_benchmark_from_module


class BenchmarkIdentifier(NamedTuple):
    module: str
    klass: str
    method: str


def split_identifier(value: str) -> BenchmarkIdentifier:
    """Split a string into a module and class identifier."""
    parts = value.split('.')
    if len(parts) != 3:
        raise argparse.ArgumentError(
            "Benchmark identifier should be in the form 'module.class.benchmark'"
        )
    return BenchmarkIdentifier(*parts)


def main():
    parser = argparse.ArgumentParser(
        description='Run selected napari benchmarks for debugging.'
    )
    parser.add_argument(
        'benchmark', type=split_identifier, help='Benchmark to run.'
    )
    args = parser.parse_args()
    module = importlib.import_module(
        f'.{args.benchmark.module}', package='napari.benchmarks'
    )
    run_benchmark_from_module(
        module, args.benchmark.klass, args.benchmark.method
    )


if __name__ == '__main__':
    main()



import numpy as np
import pytest
from qtpy.QtCore import Qt

from napari._pydantic_compat import ValidationError
from napari._vispy.utils.cursor import QtCursorVisual
from napari._vispy.utils.visual import get_view_direction_in_scene_coordinates
from napari.components._viewer_constants import CursorStyle


def test_get_view_direction_in_scene_coordinates(make_napari_viewer):
    viewer = make_napari_viewer()

    # Note: as of 0.5.6, setting the dims ndim to 3 with no layers leaves the
    # viewer in an inconsistent state, because the dims are 3 but the layers
    # extent is only 2D. Therefore, instead of setting dims to 3 we add a 3D
    # dataset to the viewer
    _ = viewer.add_image(np.random.random((2, 3, 4)))

    # reset view sets the camera angles to (0, 0, 90)
    viewer.dims.ndisplay = 3

    # get the viewbox
    view_box = viewer.window._qt_viewer.canvas.view

    # get the view direction
    view_dir = get_view_direction_in_scene_coordinates(
        view_box, viewer.dims.ndim, viewer.dims.displayed
    )
    np.testing.assert_allclose(view_dir, [-1, 0, 0], atol=1e-8)


def test_get_view_direction_in_scene_coordinates_2d(make_napari_viewer):
    """view_direction should be None in 2D"""
    viewer = make_napari_viewer()

    # reset view sets the camera angles to (0, 0, 90)
    viewer.dims.ndim = 3
    viewer.dims.ndisplay = 2

    # get the viewbox
    view_box = viewer.window._qt_viewer.canvas.view

    # get the view direction
    view_dir = get_view_direction_in_scene_coordinates(
        view_box, viewer.dims.ndim, viewer.dims.displayed
    )

    assert view_dir is None


def test_set_cursor(make_napari_viewer):
    viewer = make_napari_viewer()
    viewer.cursor.style = CursorStyle.SQUARE.value
    viewer.cursor.size = 10
    assert (
        viewer.window._qt_viewer.canvas.cursor.shape()
        == Qt.CursorShape.BitmapCursor
    )

    viewer.cursor.size = 5
    assert (
        viewer.window._qt_viewer.canvas.cursor.shape()
        == QtCursorVisual['cross'].value
    )

    viewer.cursor.style = CursorStyle.CIRCLE.value
    viewer.cursor.size = 100

    assert viewer._brush_circle_overlay.visible
    assert viewer._brush_circle_overlay.size == viewer.cursor.size

    with pytest.raises(ValidationError):
        viewer.cursor.style = 'invalid'



"""
Interaction box image
=====================

This example demonstrates activating 'transform' mode on the image layer.
This allows the user to manipulate the image via the interaction box
(blue box and points around the image).

.. tags:: experimental
"""

from skimage import data

import napari

viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True)
viewer.layers.selection.active.mode = 'transform'

if __name__ == '__main__':
    napari.run()



"""
Spherical points
================

.. tags:: experimental
"""
import numpy as np

import napari

np.random.seed()

pts = np.random.rand(100, 3) * 100
colors = np.random.rand(100, 3)
sizes = np.random.rand(100) * 20 + 10

viewer = napari.Viewer(ndisplay=3)
pts_layer = viewer.add_points(
    pts,
    face_color=colors,
    size=sizes,
    shading='spherical',
    border_width=0,
)

# antialiasing is currently a bit broken, this is especially bad in 3D so
# we turn it off here
pts_layer.antialiasing = 0

viewer.reset_view()

if __name__ == '__main__':
    napari.run()



"""
Drag and Drop Python Code Example
=================================

This example demonstrates how to execute a python script by drag'n'drop

To use this example, open napari and drag this file into the viewer.

.. tags:: interactivity
"""
from napari import Viewer
from napari.types import ImageData


def add_layers(img1: ImageData, img2: ImageData) -> ImageData:
    return img1 + img2

viewer = Viewer()

viewer.open_sample('napari', 'cells3d')
viewer.window.add_function_widget(add_layers)




"""
magicgui with threadpoolexec
============================

An example of calling a threaded function from a magicgui ``dock_widget``.

using ``ThreadPoolExecutor``
Note: this example requires python >= 3.9

.. tags:: gui
"""
from concurrent.futures import Future, ThreadPoolExecutor

from magicgui import magic_factory
from skimage import data
from skimage.feature import blob_log

import napari
from napari.types import ImageData, LayerDataTuple

pool = ThreadPoolExecutor()


@magic_factory(
    min_sigma={'min': 0.5, 'max': 15, 'step': 0.5},
    max_sigma={'min': 1, 'max': 200, 'step': 0.5},
    num_sigma={'min': 1, 'max': 20},
    threshold={'min': 0, 'max': 1000, 'step': 0.1},
)
def make_widget(
    image: ImageData,
    min_sigma: float = 5,
    max_sigma: float = 30,
    num_sigma: int = 10,
    threshold: float = 0.3,
) -> Future[LayerDataTuple]:

    # long running function
    def _make_blob():
        # skimage.feature may take a while depending on the parameters
        blobs = blob_log(
            image,
            min_sigma=min_sigma,
            max_sigma=max_sigma,
            num_sigma=num_sigma,
            threshold=threshold,
        )
        data = blobs[:, : image.ndim]
        kwargs = {
            'size': blobs[:, -1],
            'border_color': 'red',
            'border_width': 2,
            'border_width_is_relative': False,
            'face_color': 'transparent',
        }
        return (data, kwargs, 'points')

    return pool.submit(_make_blob)


viewer = napari.Viewer()
viewer.window.add_dock_widget(make_widget(), area='right')
viewer.add_image(data.hubble_deep_field().mean(-1))

napari.run()
pool.shutdown(wait=True)



"""
nD points
=========

Display one points layer on top of one 4-D image layer using the
add_points and add_image APIs, where the markes are visible as nD objects
across the dimensions, specified by their size

.. tags:: visualization-nD
"""

import numpy as np
from skimage import data

import napari

blobs = np.stack(
    [
        data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=f
        )
        for f in np.linspace(0.05, 0.5, 10)
    ],
    axis=0,
)
viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float))

# add the points
points = np.array(
    [
        [0, 0, 100, 100],
        [0, 0, 50, 120],
        [1, 0, 100, 40],
        [2, 10, 110, 100],
        [9, 8, 80, 100],
    ], dtype=float
)
viewer.add_points(
    points, size=10, face_color='blue', out_of_slice_display=True
)

if __name__ == '__main__':
    napari.run()



"""
3D vector field and image across time
=====================================

This example is from a mechanical test on a granular material, where multiple 3D volume have been acquired using x-ray tomography during axial loading.
We have used [`spam`](https://www.spam-project.dev/) to label individual particles (with watershed and some tidying up manually), and these particles are tracked through time with 3D "discrete" volume correlation *incrementally*:
t0 → t1 and then t1 → t2.
Although we're also measuring rotations (and strains), but here we're interested to visualise the displacement field on top of the image to spot tracking errors.

.. tags:: visualization-nD
"""

import numpy as np
import pooch
import tifffile

import napari

###############################################################################
# Input data
# -----------
#
# Input data are therefore:
#   - A series of 3D greyscale images
#   - A series of measured transformations between subsequent pairs of images
#   - We also have consistent labels through time which could also be visualised but are not here
#
# Let's download it!

grey_files = sorted(pooch.retrieve(
    "doi:10.5281/zenodo.17668709/grey.zip",
    known_hash="md5:760be2bad68366872111410776563760",
    processor=pooch.Unzip(),
    progressbar=True
))

# Load individual 3D images as a 4D with a list comprehension, skipping last one
# result is a T, Z, Y, X 16-bit array
greys = np.array([tifffile.imread(grey_file) for grey_file in grey_files[0:-1]])

# load incremental TSV tracking files from spam-ddic, [::2] is to skip VTK files also in folder
tracking_files = sorted(pooch.retrieve(
    "doi:10.5281/zenodo.17668709/ddic.zip",
    known_hash="md5:2d7c6a052f53b4a827ff4e4585644fac",
    processor=pooch.Unzip(),
    progressbar=True
))[::2]


###############################################################################
# Collect data together for napari
#----------------------------------
# We will loop through all the images in order to prepare the necessary data structure for napari
#
# These variables are going to contain coordinates, displacements and lengths (for colouring vectors) for all timesteps
coords_all = []
disps_all = []
lengths_all = []

for t, tracking_file in enumerate(tracking_files):
    # load the indicator for convergence
    returnStatus = np.genfromtxt(tracking_file, skip_header=1, usecols=(19))

    # Load coords and displacements, keeping only converged results (returnStatus==2)
    coords = np.genfromtxt(tracking_file, skip_header=1, usecols=(1,2,3))[returnStatus==2]
    disps = np.genfromtxt(tracking_file, skip_header=1, usecols=(4,5,6))[returnStatus==2]

    # Compute lengths in order to colour vectors
    lengths = np.linalg.norm(disps, axis=1)

    # Prepend an extra dimension to coordinates to place them in time, and fill it with the incremental t
    coords = np.hstack([np.ones((coords.shape[0],1))*t, coords])
    # Preprend zeros to the displacements (the "end" of the vector), since they do not displace through time
    disps = np.hstack([np.zeros((disps.shape[0],1)), disps])

    # Add to lists
    coords_all.append(coords)
    disps_all.append(disps)
    lengths_all.append(lengths)

# Concatenate into arrays
coords_all = np.concatenate(coords_all)
disps_all = np.concatenate(disps_all)
lengths_all = np.concatenate(lengths_all)
# Stack this into an array of size N (individual points) x 2 (vector start and length) x 4 (tzyx)
coords_displacements_all = np.stack([coords_all, disps_all], axis=1)

viewer = napari.Viewer(ndisplay=3)

viewer.add_image(
    greys,
    contrast_limits=[15000, 50000],
    rendering="attenuated_mip",
    attenuation=0.333
)

viewer.add_vectors(
  coords_displacements_all,
  vector_style='arrow',
  length=1,
  properties={'disp_norm': lengths_all},
  edge_colormap='plasma',
  edge_width=3,
  out_of_slice_display=True,
)

viewer.camera.angles = (2,-11,-23.5)
viewer.camera.orientation = ('away','up','right')
viewer.dims.current_step = (11,199, 124, 124)


if __name__ == '__main__':
    napari.run()



import csv
import itertools
import os
import re
from collections.abc import Sequence
from contextlib import contextmanager, suppress
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union
from urllib.parse import urlparse

import dask.array as da
import imageio.v3 as iio
import numpy as np
import requests
from dask import delayed

from napari.utils.misc import abspath_or_url
from napari.utils.notifications import notification_manager
from napari.utils.translations import trans

if TYPE_CHECKING:
    from napari import Viewer
    from napari.types import FullLayerData, LayerData, ReaderFunction


def _alphanumeric_key(s: str) -> list[str | int]:
    """Convert string to list of strings and ints that gives intuitive sorting."""
    return [int(c) if c.isdigit() else c for c in re.split('([0-9]+)', s)]


URL_REGEX = re.compile(r'https?://|ftps?://|file://|file:\\')

_DROPPED_SCRIPTS_NAMESPACE = {}
# This is a global dictionary to store the namespace for scripts that are
# executed using drag and drop or the Python file reader.
# The content is a mapping from the script path to the namespace.
# The dict should not be overwritten but only modified, as
# it may be broken the execution of scripts that are already running.


def _is_url(filename):
    """Return True if string is an http or ftp path.

    Originally vendored from scikit-image/skimage/io/util.py
    """
    return isinstance(filename, str) and URL_REGEX.match(filename) is not None


def _git_provider_url_to_raw_url(filename: str) -> str:
    """Convert a git provider's URL to a raw file URL.

    A git provider could be GitHub URL, GitHub Gist URL, or GitLab URL.
    Parameters
    ----------
    filename : str
        The git provider URL to convert.
    Returns
    -------
    str
        The raw file URL.
    """
    parsed_url = urlparse(filename)
    # For a GitLab file URL that contains `blob/` replace with `raw`
    if 'gitlab' in parsed_url.netloc:
        return filename.replace('blob/', 'raw/')
    # For GitHub gists, we need to substitute `githubusercontent` and
    # append `/raw` to get the raw content
    if parsed_url.netloc == 'gist.github.com':
        base_url = filename.replace(
            'gist.github.com', 'gist.githubusercontent.com'
        )
        if not base_url.endswith('/raw'):
            if '#' in base_url:
                # Split at fragment and add /raw before it
                parts = base_url.split('#')
                base_url = f'{parts[0]}/raw' + (
                    f'#{parts[1]}' if len(parts) > 1 else ''
                )
            else:
                base_url += '/raw'
        return base_url

    # For GitHub repository URLs, substitute `raw.githubusercontent.com` and `r'/refs/heads/'`
    if parsed_url.netloc == 'github.com':
        return filename.replace(
            'github.com', 'raw.githubusercontent.com'
        ).replace('/blob/', r'/refs/heads/')

    # Return filename if no match is found for a git provider
    return filename


def imread(filename: str) -> np.ndarray:
    """Dispatch reading images to imageio.v3 imread.

    Parameters
    ----------
    filename : string
        The path or URI from which to read the image.

    Returns
    -------
    data : np.ndarray
        The image data.
    """
    filename = abspath_or_url(filename)
    ext = os.path.splitext(filename)[1]

    if ext.lower() in ('.npy',):
        return np.load(filename)

    return iio.imread(str(filename))


def _guess_zarr_path(path: str) -> bool:
    """Guess whether string path is part of a zarr hierarchy."""
    return any(part.endswith('.zarr') for part in Path(path).parts)


def read_zarr_dataset(path: str):
    """Read a local or HTTP remote zarr store

    If the store is a single array, open it. If it's a group and local,
    load it as a list of arrays. For remote groups, can't traverse the hierarchy
    via HTTP, so inform the user to open an array directly.
    If it's a group of groups, open the first group and inform the user.

    Parameters
    ----------
    path : str
        Path or URL to a zarr store or directory.

    Returns
    -------
    image : array-like
        Array or list of arrays
    shape : tuple
        Shape of array or first array in list
    """
    from napari.utils.notifications import show_info

    # For local paths, zarr.open can create a DirectoryStore on a non-existent
    # path, so we check for existence first.
    if not _is_url(path) and not os.path.exists(path):
        raise FileNotFoundError(f"Path '{path}' does not exist.")

    try:
        import zarr

        store = zarr.open(path, mode='r')
    except Exception as e:
        raise ValueError(
            trans._(
                'Failed to open zarr store at {path}. Error: {error_message}',
                deferred=True,
                path=path,
                error_message=str(e),
            )
        ) from e

    # Arrays can be opened directly, local and remote
    if isinstance(store, zarr.Array):
        image = da.from_zarr(store)
        return image, image.shape

    # if we're here, it means the path wasn't a valid array, so we check if it's a valid group
    if not isinstance(store, zarr.Group):
        raise TypeError(
            trans._(
                'Unexpected zarr type: {type_}',
                deferred=True,
                type_=type(store).__name__,
            )
        )

    # Remote zarr Groups cannot be traversed over HTTP
    if _is_url(path):
        raise ValueError(
            trans._(
                'Opening remote zarr Groups is not supported. Please provide a direct URL to a zarr Array.',
                deferred=True,
            )
        )

    group_keys = sorted(store.group_keys())

    if group_keys:
        # open the first group
        group = store[group_keys[0]]

        if len(group_keys) > 1:
            # if there are multiple groups, inform the user
            other_groups = group_keys[1:]
            show_info(
                trans._(
                    'Multiple zarr Groups found in {path}. Opening group "{group}". Other groups: {other_groups}',
                    deferred=True,
                    path=path,
                    group=group_keys[0],
                    other_groups=', '.join(other_groups),
                )
            )
    else:
        # the store consists of a single group, so open it
        group = store

    array_keys = sorted(group.array_keys())
    if not array_keys:
        raise ValueError(
            trans._(
                'No arrays found in zarr group: {path}',
                deferred=True,
                path=path,
            )
        )

    # Build list of arrays from arrays in the group
    image = [da.from_zarr(group[k]) for k in array_keys]
    return image, image[0].shape


PathOrStr = Union[str, Path]


def magic_imread(
    filenames: PathOrStr | list[PathOrStr], *, use_dask=None, stack=True
):
    """Dispatch the appropriate reader given some files.

    The files are assumed to all have the same shape.

    Parameters
    ----------
    filenames : list
        List of filenames or directories to be opened.
        A list of `pathlib.Path` objects and a single filename or `Path` object
        are also accepted.
    use_dask : bool
        Whether to use dask to create a lazy array, rather than NumPy.
        Default of None will resolve to True if filenames contains more than
        one image, False otherwise.
    stack : bool
        Whether to stack the images in multiple files into a single array. If
        False, a list of arrays will be returned.

    Returns
    -------
    image : array-like
        Array or list of images
    """
    _filenames: list[str] = (
        [str(x) for x in filenames]
        if isinstance(filenames, list | tuple)
        else [str(filenames)]
    )
    if not _filenames:  # pragma: no cover
        raise ValueError('No files found')

    # replace folders with their contents
    filenames_expanded: list[str] = []
    for filename in _filenames:
        # zarr files are folders, but should be read as 1 file
        if (
            os.path.isdir(filename)
            and not _guess_zarr_path(filename)
            and not _is_url(filename)
        ):
            dir_contents = sorted(
                glob(os.path.join(filename, '*.*')), key=_alphanumeric_key
            )
            # remove subdirectories
            dir_contents_files = filter(
                lambda f: not os.path.isdir(f), dir_contents
            )
            filenames_expanded.extend(dir_contents_files)
        else:
            filenames_expanded.append(filename)

    if use_dask is None:
        use_dask = len(filenames_expanded) > 1

    if not filenames_expanded:
        raise ValueError(
            trans._(
                'No files found in {filenames} after removing subdirectories',
                deferred=True,
                filenames=filenames,
            )
        )

    # then, read in images
    images = []
    shape = None
    for filename in filenames_expanded:
        if _guess_zarr_path(filename):
            image, zarr_shape = read_zarr_dataset(filename)
            # 1D images are currently unsupported, so skip them.
            if len(zarr_shape) == 1:
                continue
            if shape is None:
                shape = zarr_shape
        else:
            if shape is None:
                image = imread(filename)
                shape = image.shape
                dtype = image.dtype
            if use_dask:
                image = da.from_delayed(
                    delayed(imread)(filename), shape=shape, dtype=dtype
                )
            elif len(images) > 0:  # not read by shape clause
                image = imread(filename)
        images.append(image)

    if not images:
        return None

    if len(images) == 1:
        image = images[0]
    elif stack:
        if use_dask:
            image = da.stack(images)
        else:
            try:
                image = np.stack(images)
            except ValueError as e:
                if 'input arrays must have the same shape' in str(e):
                    msg = trans._(
                        'To stack multiple files into a single array with numpy, all input arrays must have the same shape. Set `use_dask` to True to stack arrays with different shapes.',
                        deferred=True,
                    )
                    raise ValueError(msg) from e
                raise  # pragma: no cover
    else:
        image = images  # return a list
    return image


def _points_csv_to_layerdata(
    table: np.ndarray, column_names: list[str]
) -> 'FullLayerData':
    """Convert table data and column names from a csv file to Points LayerData.

    Parameters
    ----------
    table : np.ndarray
        CSV data.
    column_names : list of str
        The column names of the csv file

    Returns
    -------
    layer_data : tuple
        3-tuple ``(array, dict, str)`` (points data, metadata, 'points')
    """

    data_axes = [cn.startswith('axis-') for cn in column_names]
    data = np.array(table[:, data_axes]).astype('float')

    # Add properties to metadata if provided
    prop_axes = np.logical_not(data_axes)
    if column_names[0] == 'index':
        prop_axes[0] = False
    meta: dict = {}
    if np.any(prop_axes):
        meta['properties'] = {}
        for ind in np.nonzero(prop_axes)[0]:
            values = table[:, ind]
            try:
                values = np.array(values).astype('int')
            except ValueError:
                with suppress(ValueError):
                    values = np.array(values).astype('float')
            meta['properties'][column_names[ind]] = values

    return data, meta, 'points'


def _shapes_csv_to_layerdata(
    table: np.ndarray, column_names: list[str]
) -> 'FullLayerData':
    """Convert table data and column names from a csv file to Shapes LayerData.

    Parameters
    ----------
    table : np.ndarray
        CSV data.
    column_names : list of str
        The column names of the csv file

    Returns
    -------
    layer_data : tuple
        3-tuple ``(array, dict, str)`` (points data, metadata, 'shapes')
    """

    data_axes = [cn.startswith('axis-') for cn in column_names]
    raw_data = np.array(table[:, data_axes]).astype('float')

    inds = np.array(table[:, 0]).astype('int')
    n_shapes = max(inds) + 1
    # Determine when shape id changes
    transitions = list((np.diff(inds)).nonzero()[0] + 1)
    shape_boundaries = [0, *transitions] + [len(table)]
    if n_shapes != len(shape_boundaries) - 1:
        raise ValueError(
            trans._('Expected number of shapes not found', deferred=True)
        )

    data = []
    shape_type = []
    for ind_a, ind_b in itertools.pairwise(shape_boundaries):
        data.append(raw_data[ind_a:ind_b])
        shape_type.append(table[ind_a, 1])

    return data, {'shape_type': shape_type}, 'shapes'


def _guess_layer_type_from_column_names(
    column_names: list[str],
) -> str | None:
    """Guess layer type based on column names from a csv file.

    Parameters
    ----------
    column_names : list of str
        List of the column names from the csv.

    Returns
    -------
    str or None
        Layer type if recognized, otherwise None.
    """

    if {'index', 'shape-type', 'vertex-index', 'axis-0', 'axis-1'}.issubset(
        column_names
    ):
        return 'shapes'
    if {'axis-0', 'axis-1'}.issubset(column_names):
        return 'points'
    return None


def read_csv(
    filename: str, require_type: str | None = None
) -> tuple[np.ndarray, list[str], str | None]:
    """Return CSV data only if column names match format for ``require_type``.

    Reads only the first line of the CSV at first, then optionally raises an
    exception if the column names are not consistent with a known format, as
    determined by the ``require_type`` argument and
    :func:`_guess_layer_type_from_column_names`.

    Parameters
    ----------
    filename : str
        Path of file to open
    require_type : str, optional
        The desired layer type. If provided, should be one of the keys in
        ``csv_reader_functions`` or the string "any".  If ``None``, data, will
        not impose any format requirements on the csv, and data will always be
        returned.  If ``any``, csv must be recognized as one of the valid layer
        data formats, otherwise a ``ValueError`` will be raised.  If a specific
        layer type string, then a ``ValueError`` will be raised if the column
        names are not of the predicted format.

    Returns
    -------
    (data, column_names, layer_type) : Tuple[np.array, List[str], str]
        The table data and column names from the CSV file, along with the
        detected layer type (string).

    Raises
    ------
    ValueError
        If the column names do not match the format requested by
        ``require_type``.
    """
    with open(filename, newline='') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        column_names = next(reader)

        layer_type = _guess_layer_type_from_column_names(column_names)
        if require_type:
            if not layer_type:
                raise ValueError(
                    trans._(
                        'File "{filename}" not recognized as valid Layer data',
                        deferred=True,
                        filename=filename,
                    )
                )
            if layer_type != require_type and require_type.lower() != 'any':
                raise ValueError(
                    trans._(
                        'File "{filename}" not recognized as {require_type} data',
                        deferred=True,
                        filename=filename,
                        require_type=require_type,
                    )
                )

        data = np.array(list(reader))
    return data, column_names, layer_type


csv_reader_functions = {
    'points': _points_csv_to_layerdata,
    'shapes': _shapes_csv_to_layerdata,
}


def csv_to_layer_data(
    path: str, require_type: str | None = None
) -> Optional['FullLayerData']:
    """Return layer data from a CSV file if detected as a valid type.

    Parameters
    ----------
    path : str
        Path of file to open
    require_type : str, optional
        The desired layer type. If provided, should be one of the keys in
        ``csv_reader_functions`` or the string "any".  If ``None``,
        unrecognized CSV files will simply return ``None``.  If ``any``,
        unrecognized CSV files will raise a ``ValueError``.  If a specific
        layer type string, then a ``ValueError`` will be raised if the column
        names are not of the predicted format.

    Returns
    -------
    layer_data : tuple, or None
        3-tuple ``(array, dict, str)`` (points data, metadata, layer_type) if
        CSV is recognized as a valid type.

    Raises
    ------
    ValueError
        If ``require_type`` is not ``None``, but the CSV is not detected as a
        valid data format.
    """
    try:
        # pass at least require "any" here so that we don't bother reading the
        # full dataset if it's not going to yield valid layer_data.
        _require = require_type or 'any'
        table, column_names, _type = read_csv(path, require_type=_require)
    except ValueError:
        if not require_type:
            return None
        raise
    if _type in csv_reader_functions:
        return csv_reader_functions[_type](table, column_names)
    return None  # only reachable if it is a valid layer type without a reader


def _csv_reader(path: str | Sequence[str]) -> list['LayerData']:
    if isinstance(path, str):
        layer_data = csv_to_layer_data(path, require_type=None)
        return [layer_data] if layer_data else []
    return [
        layer_data
        for p in path
        if (layer_data := csv_to_layer_data(p, require_type=None))
    ]


def _magic_imreader(path: str) -> list['LayerData']:
    return [(magic_imread(path),)]


def napari_get_reader(
    path: str | list[str],
) -> 'ReaderFunction':
    """Our internal fallback file reader at the end of the reader plugin chain.

    This will assume that the filepath is an image, and will pass all of the
    necessary information to viewer.add_image().

    Parameters
    ----------
    path : str
        path to file/directory

    Returns
    -------
    callable
        function that returns layer_data to be handed to viewer._add_layer_data
    """
    if isinstance(path, str) and path.endswith('.csv'):
        return _csv_reader

    return _magic_imreader


@contextmanager
def _patch_viewer_new():
    """Context manager to patch the viewer's new method."""
    from napari.viewer import Viewer, current_viewer

    original_new = Viewer.__new__
    original_init = Viewer.__init__

    def patched_init(self, *args, **kwargs):
        Viewer.__init__ = original_init

    def patched_new(cls, *args, **kwargs):
        ndisplay = None
        if len(kwargs) == 1 and 'ndisplay' in kwargs:
            ndisplay = kwargs.pop('ndisplay')

        if not kwargs and not args:
            viewer = current_viewer()
            if ndisplay is not None:
                viewer.dims.ndisplay = ndisplay
            if viewer is not None:
                Viewer.__new__ = original_new
                return viewer
        Viewer.__init__ = original_init
        return original_new(cls)

    Viewer.__new__ = patched_new
    Viewer.__init__ = patched_init
    try:
        yield
    finally:
        Viewer.__new__ = original_new
        Viewer.__init__ = original_init


@contextmanager
def _patch_napari_run():
    """Context manager to patch napari.run to always be a no-op.

    napari.run() executes the Qt event loop, *except* when napari
    is running in IPython and therefore IPython's Qt integration
    already has the event loop.

    When running a script by dragging-and-dropping onto a
    running napari Viewer, we already have an event loop, so we
    should not start a new nested loop, even though we are not
    in IPython.

    This context manager temporarily patches the IPython check
    to always return True, causing a fast exit from napari.run()
    without a new event loop.
    """
    from napari._qt import qt_event_loop

    original_ipython_check = qt_event_loop._ipython_has_eventloop

    def patched_ipython_check() -> bool:
        """A patched ipython_check that always returns True.

        napari's script running from drag-and-dropping a script
        into a viewer uses this patch to prevent nested event loops.
        """
        return True

    qt_event_loop._ipython_has_eventloop = patched_ipython_check
    try:
        yield
    finally:
        qt_event_loop._ipython_has_eventloop = original_ipython_check


def filter_variables(variables: dict[str, Any]) -> dict[str, Any]:
    res = variables.copy()
    res.pop('viewer', None)
    res.pop(
        '__name__', None
    )  # Remove the __name__ variable to not affect the console
    return res


def _add_dropped_scripts_to_console(
    variables: dict[str, Any], viewer: 'Viewer | None'
) -> None:
    if viewer is None:
        return

    variables = filter_variables(variables)

    if viewer.window._qt_viewer._console is None:
        viewer.window._qt_viewer.add_to_console_backlog(variables)
    else:
        console = viewer.window._qt_viewer._console
        console.push(variables)


def load_and_execute_python_code(script_path: str) -> list['LayerData']:
    """Load and execute Python code from a file.

    Parameters
    ----------
    script_path : str
        Path to the Python file to be executed.
    """
    if _is_url(script_path):
        # download the script from the URL

        response = requests.get(_git_provider_url_to_raw_url(script_path))
        response.raise_for_status()
        code = response.text
    else:
        code = Path(script_path).read_text()
    execute_python_code(code, script_path)
    return [(None,)]


def execute_python_code(code: str, script_path: str | Path) -> None:
    """Execute Python code in the current viewer's context.

    Store the executed cod variables in _DROPPED_SCRIPTS_NAMESPACE dict

    Parameters
    ----------
    code: str
        python code to be executed
    script_path: str | Path
        Path to the script file from which the code is executed.
        Used to store the namespace in the _DROPPED_SCRIPTS_NAMESPACE.
    """
    from napari.viewer import current_viewer

    with _patch_viewer_new(), _patch_napari_run():
        try:
            viewer = current_viewer()
            script_namespace = _DROPPED_SCRIPTS_NAMESPACE.setdefault(
                script_path, {}
            )
            # The `__name__` variable is storing the name of the module.
            # If a module is imported, it is set to the module name.
            # If a module is executed with `python -m ...` or
            # `python script.py` it is set to '__main__'.
            # If code is executed with `exec(code, namespace)` it is set to `builtins` if
            # `__name__` is not set in the namespace.
            # So ww set it to `__main__` to execute `if __name__ == '__main__':` blocks
            script_namespace['__name__'] = '__main__'
            exec(code, script_namespace)
            _add_dropped_scripts_to_console(
                _DROPPED_SCRIPTS_NAMESPACE[script_path], viewer
            )
        except BaseException as e:  # noqa: BLE001
            notification_manager.receive_error(type(e), e, e.__traceback__)


def napari_get_py_reader(path: str) -> 'ReaderFunction | None':
    """Return a reader function for Python files.

    This function is used to read Python files and execute their content.
    It returns a callable that executes the code in the file.

    Parameters
    ----------
    path : str
        Path to the Python file to be executed.

    Returns
    -------
    callable
        A function that executes the Python code in the specified file.
    """
    if not os.path.exists(path) and not _is_url(path):
        return None
    if os.path.splitext(path)[1] != '.py':
        return None
    return load_and_execute_python_code



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any
from warnings import warn

import numpy as np
import pandas as pd
from qtpy.QtCore import (
    QAbstractTableModel,
    QItemSelection,
    QItemSelectionModel,
    QModelIndex,
    QSortFilterProxyModel,
    Qt,
    QTimer,
)
from qtpy.QtGui import (
    QGuiApplication,
    QKeySequence,
)
from qtpy.QtWidgets import (
    QAbstractItemView,
    QComboBox,
    QFileDialog,
    QLabel,
    QPushButton,
    QStyledItemDelegate,
    QTableView,
    QVBoxLayout,
    QWidget,
)
from superqt import QToggleSwitch

from napari.utils.history import get_save_history
from napari.utils.misc import in_ipython

if TYPE_CHECKING:
    import napari
    import napari.components


class PandasModel(QAbstractTableModel):
    """Qt Model for a pandas DataFrame.

    Follows the Qt Model/View protocol, implementing all the needed methods
    for a QTableView to be able to read/write changes to a pandas dataframe.
    Columns of specific dtypes get special treatment (bools become tickboxes,
    categoricals are exposed through comboboxes, etc).
    It's designed to be used in conjunction with the BoolFriendlyProxyModel
    in order to properly sort, and with the DelegateCategorical in order to
    provide comboboxes for categoricals.

    This model supports the concept of immutable columns identified by name, not position.
    For pandas DataFrames, column 0 (representing the DataFrame index) is automatically
    treated as immutable.

    Parameters
    ----------
    df : pd.DataFrame, optional
        The pandas DataFrame to wrap.
    immutable_columns : list[str], optional
        List of column names to be treated as immutable (read-only).
    parent : QObject, optional
        Parent Qt object.
    """

    def __init__(
        self,
        df: pd.DataFrame | None = None,
        immutable_columns: list[str] | None = None,
        source_column_name: str | None = None,
        parent=None,
    ):
        super().__init__(parent)
        self.df = df if df is not None else pd.DataFrame()
        self.editable = False
        self._immutable_columns = set(immutable_columns or [])
        self._source_column_name = source_column_name
        self._original_categories: dict[
            str, dict[str, list]
        ] = {}  # {layer_name: {col_name: [categories]}}
        self._add_index_to_immutable()

    # model methods necessary for qt
    def rowCount(self, parent=None):
        return self.df.shape[0]

    def columnCount(self, parent=None):
        return self.df.shape[1] + 1  # include index

    def data(self, index: QModelIndex, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():  # pragma: no cover
            return None

        row = index.row()
        col = index.column()
        value, dtype = self._get_cell_value_and_dtype(row, col)

        # show booleans as respective checkboxes
        if (
            role == Qt.ItemDataRole.CheckStateRole
            and pd.api.types.is_bool_dtype(dtype)
        ):
            return Qt.CheckState.Checked if value else Qt.CheckState.Unchecked

        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            if pd.api.types.is_float_dtype(dtype):
                return float(value)
            if pd.api.types.is_integer_dtype(dtype):
                return int(value)
            if isinstance(
                value, pd.Timestamp
            ) or pd.api.types.is_datetime64_any_dtype(dtype):
                return value.strftime('%Y-%m-%d')
            if pd.api.types.is_bool_dtype(dtype):
                if role == Qt.ItemDataRole.DisplayRole:
                    return ''  # do not show True/False text
                if role == Qt.ItemDataRole.EditRole:
                    return bool(value)  # needed for proper sorting
            if value is pd.NA:
                return pd.NA  # shows empty cell in the view
            return str(value)

        return None

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role=Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        if isinstance(self.df, pd.DataFrame):
            if orientation == Qt.Orientation.Horizontal:
                if self._has_source_column():
                    # With source column: col 0 = Source, col 1 = Index, col 2+ = features
                    if section == 0:
                        return self._source_column_name
                    if section == 1:
                        return self.df.index.name or 'Index'
                    return self.df.columns[section - 1]  # Offset by 1
                # Without source column: col 0 = Index, col 1+ = features
                if section == 0:
                    return self.df.index.name or 'Index'
                return self.df.columns[section - 1]
            # Vertical header
            return self.df.index[section]
        # TODO: For non-pandas dataframes, implement appropriate header handling
        return str(section)

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled

        row = index.row()
        col = index.column()
        value, dtype = self._get_cell_value_and_dtype(row, col)

        # Check if this column is immutable
        flags = Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        if self.is_column_immutable(col):
            return flags

        # Cells with pd.NA are not editable
        if value is pd.NA:
            return flags

        if self.editable:
            # make boolean columns checkable
            if pd.api.types.is_bool_dtype(dtype):
                flags |= Qt.ItemFlag.ItemIsUserCheckable
            else:
                flags |= Qt.ItemFlag.ItemIsEditable

        return flags

    def setData(
        self, index: QModelIndex, value: Any, role=Qt.ItemDataRole.EditRole
    ) -> bool:
        if not index.isValid():
            return False

        col = index.column()
        # Check if this column is immutable
        if self.is_column_immutable(col):
            return False

        row = index.row()
        # Get DataFrame column index
        df_col = self._get_df_col_idx(col)
        if df_col is None:
            # Index or Layer column - should be caught by immutability check
            return False
        dtype = np.dtype(type(self.df.iat[row, df_col]))

        # checkboxes
        if (
            role == Qt.ItemDataRole.CheckStateRole
            and pd.api.types.is_bool_dtype(dtype)
        ):
            self.df.iat[row, df_col] = (
                Qt.CheckState(value) == Qt.CheckState.Checked
            )
            self.dataChanged.emit(
                index, index, [Qt.ItemDataRole.CheckStateRole]
            )
            return True

        if role == Qt.ItemDataRole.EditRole:
            # Warn if setting a string cell to a number-like value
            if pd.api.types.is_string_dtype(dtype) and self._looks_like_number(
                str(value)
            ):
                warn(
                    f'Setting a string cell to the value "{value}" which looks like a number.',
                    UserWarning,
                    stacklevel=3,
                )

            try:
                if not isinstance(dtype, pd.CategoricalDtype):
                    value = dtype.type(value)
                self.df.iat[row, df_col] = value
            except (TypeError, ValueError):
                # Type error is for categorical types that cannot be converted
                # Value error is for invalid values in dtype.type(value)
                return False
            self.dataChanged.emit(
                index,
                index,
                [Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole],
            )
            return True

        return False

    # custom methods
    @contextmanager
    def changing(self):
        self.layoutAboutToBeChanged.emit()
        yield
        self.layoutChanged.emit()

    def replace_data(self, df):
        with self.changing():
            self.df = df
            self._add_index_to_immutable()
            # Reset provenance when data is replaced
            self._original_categories = {}

    def set_original_categories(
        self, categories_dict: dict[str, dict[str, list]]
    ) -> None:
        """Store original categories by layer before concat conversion to object dtype.

        Parameters
        ----------
        categories_dict : dict[str, dict[str, list]]
            Mapping of layer_name -> {column_name -> [category_values]}
            Only includes columns that were categorical in each layer.
        """
        self._original_categories = categories_dict

    def is_categorical_in_layer(self, col_name: str, layer_name: str) -> bool:
        """Check if a column was categorical in a specific layer.

        Parameters
        ----------
        col_name : str
            Column name to check
        layer_name : str
            Layer name to check against

        Returns
        -------
        bool
            True if column was categorical in that layer before concat
        """
        return col_name in self._original_categories.get(layer_name, {})

    def get_original_categories(
        self, col_name: str, layer_name: str
    ) -> list | None:
        """Get original categories for a column in a specific layer.

        Parameters
        ----------
        col_name : str
            Column name
        layer_name : str
            Layer name

        Returns
        -------
        list | None
            List of categories if column was categorical in that layer, else None
        """
        return self._original_categories.get(layer_name, {}).get(col_name)

    def set_immutable_columns(self, column_names: list[str]):
        """Set columns that should be treated as immutable (read-only).

        Parameters
        ----------
        column_names : list[str]
            List of column names to mark as immutable. The index column (0)
            is always immutable regardless of this setting.
        """
        self._immutable_columns = set(column_names)
        self._add_index_to_immutable()

    def get_immutable_columns(self) -> list[str]:
        """Get the list of column names that are marked as immutable.

        Returns
        -------
        list[str]
            List of column names that are immutable (read-only).
        """
        return list(self._immutable_columns)

    def _add_index_to_immutable(self):
        """Add the index name (or 'Index') to the immutable columns if a pandas DataFrame."""
        if isinstance(self.df, pd.DataFrame):
            if self.df.index.name is not None:
                self._immutable_columns.add(self.df.index.name)
            else:
                self._immutable_columns.add('Index')

    def _has_source_column(self) -> bool:
        """Check if the DataFrame has a source identifier column.

        Returns
        -------
        bool
            True if source column exists in the DataFrame.
        """
        return (
            self._source_column_name is not None
            and self._source_column_name in self.df.columns
        )

    def _get_df_col_idx(self, view_col: int) -> int | None:
        """Convert view column index to DataFrame column index.

        When a source column exists, the view layout is:
        - view col 0 → source column (e.g., 'Layer') at df.columns[0]
        - view col 1 → DataFrame index
        - view col 2+ → DataFrame columns starting at df.columns[1]

        Without a source column, the view layout is:
        - view col 0 → DataFrame index
        - view col 1+ → DataFrame columns starting at df.columns[0]

        Parameters
        ----------
        view_col : int
            The column index from the view

        Returns
        -------
        int | None
            DataFrame column index, or None if this is the index column or source column
        """
        if self._has_source_column():
            # Multi-source mode: Source, Index, then features
            if view_col == 0:
                return (
                    None  # Source column (special handling needed by caller)
                )
            if view_col == 1:
                return None  # Index column
            return view_col - 1  # view col 2 → df col 1, etc.
        # Single-source mode: Index, then features
        if view_col == 0:
            return None  # Index column
        return view_col - 1  # view col 1 → df col 0, etc.

    def _get_cell_value_and_dtype(
        self, row: int, col: int
    ) -> tuple[Any, np.dtype]:
        """Get value and dtype for a cell at the given row and column.

        Parameters
        ----------
        row : int
            Row index in the model
        col : int
            Column index in the model (view column)

        Returns
        -------
        tuple[Any, np.dtype]
            The cell value and its dtype
        """
        if self._has_source_column():
            # With source column: col 0 = Source, col 1 = Index, col 2+ = features
            if col == 0:
                value = self.df.iat[row, 0]  # Source column is first in df
            elif col == 1:
                value = self.df.index[row]
            else:
                value = self.df.iat[
                    row, col - 1
                ]  # Offset by 1 (source column)
        else:
            # Without source column: col 0 = Index, col 1+ = features
            if col == 0:
                value = self.df.index[row]
            else:
                value = self.df.iat[row, col - 1]

        return value, np.dtype(type(value))

    def is_column_immutable(self, col_idx: int) -> bool:
        """Check if a column is immutable based on its index.

        For pandas DataFrames:
        - With source column: columns 0 (source) and 1 (Index) are immutable
        - Without source column: column 0 (Index) is immutable
        Columns with names in self._immutable_columns are also immutable.

        Parameters
        ----------
        col_idx : int
            Column index to check

        Returns
        -------
        bool
            True if the column is immutable, False otherwise
        """
        if not isinstance(self.df, pd.DataFrame):
            return False

        # Check if this is an immutable structural column (source or index)
        if self._has_source_column():
            # With source: Source (col 0) and Index (col 1) are immutable
            if col_idx in (0, 1):
                return True
        else:
            # Without source: Index (col 0) is immutable
            if col_idx == 0:
                return True

        # Check if the feature column name is in immutable set
        if col_idx - 1 < len(self.df.columns):
            col_name = self.df.columns[col_idx - 1]
            return col_name in self._immutable_columns

        return False

    @staticmethod
    def _looks_like_number(value: str) -> bool:
        """Check if a string looks like a number (int or float).

        Parameters
        ----------
        value : str
            The string value to check

        Returns
        -------
        bool
            True if the string looks like an int or float, False otherwise
        """
        if not isinstance(value, str):
            return False
        stripped = value.strip()
        if not stripped:
            return False
        try:
            float(stripped)
        except ValueError:
            return False
        else:
            return True


class DelegateCategorical(QStyledItemDelegate):
    """Delegate that uses comboboxes as editors for categorical data."""

    def createEditor(self, parent, option, index: QModelIndex) -> QWidget:
        proxy_model = index.model()
        source_model = proxy_model.sourceModel()
        source_index = proxy_model.mapToSource(index)
        col = source_index.column()

        # Get value via proxy_model for EditRole
        value = proxy_model.data(index, Qt.ItemDataRole.EditRole)
        dtype = np.dtype(type(value))
        # If bool dtype or value is bool, let Qt use default checkbox editor
        if pd.api.types.is_bool_dtype(dtype) or isinstance(value, bool):
            return None

        # Map view column to DataFrame column name
        df_col_idx = source_model._get_df_col_idx(col)
        if df_col_idx is None:
            # Index or Layer column - no editor needed
            return None
        col_name = source_model.df.columns[df_col_idx]

        # Determine layer name for this row
        if source_model._has_source_column():
            # Multi-source case: get source from the source column
            layer_name = source_model.df.iloc[source_index.row()][
                source_model._source_column_name
            ]
        else:
            # Single-layer case: get the layer name from provenance (should be exactly one)
            layer_names = list(source_model._original_categories.keys())
            layer_name = layer_names[0] if layer_names else None

        # Check provenance: only create combobox if column was originally categorical
        original_categories = None
        if layer_name is not None:
            original_categories = source_model.get_original_categories(
                col_name, layer_name
            )

        if original_categories is not None:
            editor = QComboBox(parent)
            editor.addItems([str(c) for c in original_categories])
            # allow arrow keys selection
            editor.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

            # force editor to open on first click, otherwise we need 2 clicks
            QTimer.singleShot(0, editor.showPopup)
            return editor

        # Fallback: check current dtype for backward compatibility
        if isinstance(dtype, pd.CategoricalDtype):
            editor = QComboBox(parent)
            categories = source_model.df.iloc[:, df_col_idx].cat.categories
            editor.addItems([str(c) for c in categories])
            # allow arrow keys selection
            editor.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

            # force editor to open on first click, otherwise we need 2 clicks
            QTimer.singleShot(0, editor.showPopup)
            return editor

        # If float, use spinbox
        if pd.api.types.is_float_dtype(dtype):
            editor = super().createEditor(parent, option, index)
            editor.setDecimals(10)
            return editor

        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: QWidget, index: QModelIndex):
        if isinstance(editor, QComboBox):
            value = index.model().data(index, Qt.ItemDataRole.EditRole)
            i = editor.findText(value)
            if i >= 0:
                editor.setCurrentIndex(i)
        else:
            super().setEditorData(editor, index)

    def setModelData(
        self, editor: QWidget, model: QSortFilterProxyModel, index: QModelIndex
    ):
        if isinstance(editor, QComboBox):
            source_index = model.mapToSource(index)
            source_model = model.sourceModel()
            source_model.setData(
                source_index, editor.currentText(), Qt.ItemDataRole.EditRole
            )
        else:
            super().setModelData(editor, model, index)


class BoolFriendlyProxyModel(QSortFilterProxyModel):
    """Sort proxy model that handles booleans correctly and sorts immutable columns as integers if possible."""

    def lessThan(self, left: Any, right: Any) -> bool:
        left_data = self.sourceModel().data(left, Qt.ItemDataRole.EditRole)
        right_data = self.sourceModel().data(right, Qt.ItemDataRole.EditRole)

        # ensure booleans compare as expected. Not sure what happens internally in qt
        # that doesn't work, but doing it ourselves in python works.
        # One thing that breaks it are numpy bools
        if all(pd.api.types.is_bool(d) for d in (left_data, right_data)):
            return bool(left_data < right_data)

        return super().lessThan(left, right)


class PandasView(QTableView):
    """View for PandasModel to interact with pandas dataframes via a table.

    This class is designed to work in tandem with PandasModel, and allows
    the following things:
    - sorting the table without affecting the data
    - copy/pasting cells
    - edting the data with specific editors depending on the dtype
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # view/model setup with a proxy for sorting and filtering
        proxy_model = BoolFriendlyProxyModel()
        proxy_model.setSourceModel(PandasModel(source_column_name='Layer'))
        proxy_model.setSortCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.setModel(proxy_model)
        self.setSortingEnabled(True)
        # do not auto sort using index on startup
        self.horizontalHeader().setSortIndicator(
            -1, Qt.SortOrder.AscendingOrder
        )
        # disable vertical header (since we duplicate it as a column)
        self.verticalHeader().setVisible(False)
        # delegate which provides comboboxes for editing categorical values
        self.setItemDelegate(DelegateCategorical())
        # enable selection and editing
        self.setSelectionBehavior(QTableView.SelectionBehavior.SelectItems)
        self.setSelectionMode(QTableView.SelectionMode.ExtendedSelection)
        self.setEditTriggers(QAbstractItemView.EditTrigger.AllEditTriggers)

    def keyPressEvent(self, event):
        if event.matches(QKeySequence.StandardKey.Copy):
            self.copySelection()
        elif event.matches(QKeySequence.StandardKey.Paste):
            self.pasteSelection()
        else:
            super().keyPressEvent(event)

    def copySelection(self):
        selection = self.selectedIndexes()
        if not selection:
            return  # pragma: no cover

        proxy_model = self.model()
        model = proxy_model.sourceModel()
        map_func = proxy_model.mapToSource

        # determine selected block (excluding immutable columns)
        model = self.model().sourceModel()

        # Filter out immutable columns
        rows = sorted(
            {
                idx.row()
                for idx in selection
                if not model.is_column_immutable(idx.column())
            }
        )
        cols = sorted(
            {
                idx.column()
                for idx in selection
                if not model.is_column_immutable(idx.column())
            }
        )

        if not rows or not cols:
            return  # pragma: no cover

        # convert proxy indices to actual row/column labels
        actual_rows = [map_func(proxy_model.index(r, 1)).row() for r in rows]
        # Map view columns to df columns, filtering out immutable columns (None)
        actual_cols = [model._get_df_col_idx(c) for c in cols]
        actual_cols = [c for c in actual_cols if c is not None]

        sub_df = model.df.iloc[actual_rows, actual_cols]

        tsv = sub_df.to_csv(sep='\t', index=False, header=False)
        QGuiApplication.clipboard().setText(tsv)

    def pasteSelection(self):
        proxy_model = self.model()
        model = proxy_model.sourceModel()
        map_func = proxy_model.mapToSource
        df = model.df

        clipboard = QGuiApplication.clipboard().text()
        sel = self.selectedIndexes()
        # if any immutable column is in the selection, just get out
        if (
            not clipboard
            or not sel
            or not model.editable
            or any(model.is_column_immutable(s.column()) for s in sel)
        ):
            return  # pragma: no cover

        rows = clipboard.strip().split('\n')
        data = [row.split('\t') for row in rows]
        n_rows = len(data)
        n_cols = max(len(row) for row in data)

        # copy from top-left selected cell
        start_proxy_index = min(sel, key=lambda x: (x.row(), x.column()))
        start_source_index = map_func(start_proxy_index)
        start_row = start_source_index.row()
        start_col = start_source_index.column()

        # Track max modified column
        max_modified_col = start_col

        for i in range(n_rows):
            for j in range(n_cols):
                r = start_row + i
                c = start_col + j
                # Get DataFrame column index from source model column
                df_col = model._get_df_col_idx(c)
                if df_col is None:  # Index or Layer column - skip
                    continue
                if r >= df.shape[0] or df_col >= df.shape[1]:
                    continue
                val = data[i][j]
                dtype = df.dtypes.iloc[df_col]
                try:
                    if not isinstance(dtype, pd.CategoricalDtype):
                        val = dtype.type(val)
                    df.iat[r, df_col] = val
                    # Track the max column that was actually modified
                    max_modified_col = max(max_modified_col, c)
                except (ValueError, TypeError):  # pragma: no cover
                    # TODO: warning? undo?
                    pass

        # notify the model/view
        top_left = model.index(start_row, start_col)
        bottom_right = model.index(start_row + n_rows - 1, max_modified_col)
        model.dataChanged.emit(top_left, bottom_right)


class FeaturesTable(QWidget):
    """Widget to display layer features as an editable table.

    The table automatically shows the contents of the current active layer
    in the layerlist, if that layer has a features table.
    Features can be edited directly if the `editable` toggle is enabled.
    Columns can be sorted without affecting the layer data.
    Selecting a row in the table will select the corresponding data in the
    layer, and viceversa.
    Data can be copy/pasted as csv, and can be saved to file.
    """

    def __init__(
        self,
        viewer: napari.viewer.ViewerModel,
    ) -> None:
        super().__init__()
        self._selected_layers = []
        self._selection_blocked = False

        self.viewer = viewer
        self.viewer.layers.selection.events.changed.connect(
            self._on_layer_selection_change
        )
        self.viewer.layers.events.renamed.connect(self._on_layer_renamed)

        self.setLayout(QVBoxLayout())

        self.info = QLabel('')
        self.toggle = QToggleSwitch('editable')
        self.join_toggle = QToggleSwitch('shared columns only')
        self.save = QPushButton('Save as CSV...')
        self.table = PandasView()
        self.layout().addWidget(self.info)
        self.layout().addWidget(self.toggle)
        self.layout().addWidget(self.join_toggle)
        self.layout().addWidget(self.save)
        self.layout().addWidget(self.table)
        self.layout().addStretch()

        self.toggle.toggled.connect(self._on_editable_change)
        self.join_toggle.toggled.connect(self._on_join_change)
        self.save.clicked.connect(self._on_save_clicked)

        self.table.selectionModel().selectionChanged.connect(
            self._on_table_selection_changed
        )

        # Connect to dataChanged to sync edits back to layers
        self.table.model().sourceModel().dataChanged.connect(
            self._on_table_data_changed
        )

        self._on_layer_selection_change()
        self._on_editable_change()

    def _on_layer_renamed(self, event):
        """Update table and info label when a layer is renamed."""
        if not self._selected_layers:
            return

        # Update info label with new layer names
        if len(self._selected_layers) == 1:
            self.info.setText(f'Features of "{self._selected_layers[0].name}"')
        else:
            layer_names = ', '.join(
                f'"{layer.name}"'
                for layer in sorted(
                    self._selected_layers, key=lambda lyr: lyr.name
                )
            )
            self.info.setText(f'Features of [{layer_names}]')

        # Update the Layer column in the table if present
        model = self.table.model().sourceModel()
        df = model.df
        if 'Layer' in df.columns and len(self._selected_layers) > 1:
            # Rebuild the table to reflect the renamed layer
            self._on_features_change()

    @staticmethod
    def _get_selection_event_for_layer(layer):
        if hasattr(layer, 'selected_label'):
            return layer.events.selected_label
        if hasattr(layer, 'selected_data'):
            # Points layer has selected_data.events, but Shapes layer uses highlight event
            if hasattr(layer.selected_data, 'events'):
                return layer.selected_data.events
            if hasattr(layer.events, 'highlight'):
                return layer.events.highlight
        # Return None if layer doesn't have expected selection attributes
        return None

    def _disconnect_layer_events(self, layers):
        """Disconnect features and selection events from the update table callbacks."""
        for layer in layers:
            if hasattr(layer, 'events') and hasattr(layer.events, 'features'):
                layer.events.features.disconnect(self._on_features_change)
            selection_event = self._get_selection_event_for_layer(layer)
            if selection_event is not None:
                selection_event.disconnect(self._update_table_selected_cells)

    def _connect_layer_events(self, layers):
        """Connect features and selection events to the appropriate update table callbacks."""
        for layer in layers:
            if hasattr(layer, 'events') and hasattr(layer.events, 'features'):
                layer.events.features.connect(self._on_features_change)
            selection_event = self._get_selection_event_for_layer(layer)
            if selection_event is not None:
                selection_event.connect(self._update_table_selected_cells)

    def _on_layer_selection_change(self):
        """Update the table when the layer selection changes and handles event connections."""
        old_layer_list = self._selected_layers
        # Filter to only keep layers with features
        self._selected_layers = [
            layer
            for layer in self.viewer.layers.selection
            if hasattr(layer, 'features')
        ]
        if len(old_layer_list) > 0:
            self._disconnect_layer_events(old_layer_list)

        if len(self._selected_layers) > 0:
            self._connect_layer_events(self._selected_layers)

            # Show widgets and update table
            self._on_features_change()
            self.toggle.setVisible(True)
            self.save.setVisible(True)
            self.table.setVisible(True)
            self.join_toggle.setVisible(len(self._selected_layers) > 1)
            if len(self._selected_layers) == 1:
                self.info.setText(
                    f'Features of "{self._selected_layers[0].name}"'
                )
            else:
                layer_names = ', '.join(
                    f'"{layer.name}"'
                    for layer in sorted(
                        self._selected_layers, key=lambda lyr: lyr.name
                    )
                )
                self.info.setText(f'Features of [{layer_names}]')
        else:
            # Hide widgets and show appropriate message
            self.toggle.setVisible(False)
            self.join_toggle.setVisible(False)
            self.save.setVisible(False)
            self.table.setVisible(False)

            # Determine message based on original selection
            if len(self.viewer.layers.selection) > 0:
                self.info.setText('Selected layers do not have features.')
            else:
                self.info.setText('No layer selected.')

    def _on_features_change(self):
        """Update the table with the features of the currently selected layers."""
        # TODO: optimize for smaller changes?
        join_type = 'inner' if self.join_toggle.isChecked() else 'outer'

        # Extract categories before concat converts them to object dtype
        original_categories = self._get_original_categories_from_layers()

        df = self._build_multilayer_features_table(join=join_type)

        # Replace data and configure immutable columns
        model = self.table.model().sourceModel()
        model.replace_data(df)
        model.set_original_categories(original_categories)
        model.set_immutable_columns(['Layer'] if 'Layer' in df.columns else [])

        self.table.resizeColumnsToContents()
        self._update_table_selected_cells()

    def _build_multilayer_features_table(
        self, join: str = 'outer'
    ) -> pd.DataFrame:
        """Builds a features table for multiple layers."""
        df_list = []
        presence_mask_list = []
        for layer in self._selected_layers:
            # All layers in self._selected_layers are guaranteed to have features
            if layer.features is not None:
                if isinstance(layer.features, pd.DataFrame):
                    # Make a copy to avoid modifying the original layer.features
                    df = layer.features.copy()

                    # Only add 'Layer' column for multiple layers
                    if (
                        len(self._selected_layers) > 1
                        and 'Layer' not in df.columns
                    ):
                        df['Layer'] = layer.name
                        df['Layer'] = df['Layer'].astype('category')
                        # Move 'Layer' to the first column
                        cols = list(df.columns)
                        cols.remove('Layer')
                        cols.insert(0, 'Layer')
                        df = df[cols]
                    df_list.append(df)
                    presence_mask_list.append(~df.isna())
                else:
                    # TODO: Handle non-pandas dataframe libraries here
                    pass
        # Combine all dataframes and presence masks, preserving original indices per layer
        df = pd.concat(df_list, ignore_index=False, join=join)
        presence_mask = pd.concat(
            presence_mask_list, ignore_index=False, join=join
        )
        nan_cells_introduced_by_join = df.isna() & presence_mask.isna()
        # Convert columns with join-introduced NaNs to object dtype to ensure
        # pd.NA will be set instead of internally converted to NaT or np.nan
        cols_with_join_nans = nan_cells_introduced_by_join.any()
        for col in df.columns[cols_with_join_nans]:
            df[col] = df[col].astype('object')
        # Replace only NaNs introduced by outer join with pd.NA
        df = df.mask(nan_cells_introduced_by_join, pd.NA)
        return df

    def _on_editable_change(self):
        self.table.model().sourceModel().editable = self.toggle.isChecked()

    def _on_join_change(self):
        """Update the table when join mode changes."""
        self._on_features_change()

    def _get_original_categories_from_layers(
        self,
    ) -> dict[str, dict[str, list]]:
        """Extract categories by layer before concat loses dtype information.

        When concatenating DataFrames with columns that are categorical in some
        layers but strings (or missing) in others, pandas converts all to object dtype.
        This method captures the original categorical information per layer.

        Returns
        -------
        dict[str, dict[str, list]]
            Mapping: {layer_name: {column_name: [category_values]}}
            Only includes columns that were categorical in each layer.
        """
        categories_dict: dict[str, dict[str, list]] = {}
        for layer in self._selected_layers:
            if layer.features is not None and isinstance(
                layer.features, pd.DataFrame
            ):
                layer_categories: dict[str, list] = {}
                for col in layer.features.columns:
                    if isinstance(
                        layer.features[col].dtype, pd.CategoricalDtype
                    ):
                        layer_categories[col] = list(
                            layer.features[col].cat.categories
                        )
                if layer_categories:  # Only add if there are categoricals
                    categories_dict[layer.name] = layer_categories
        return categories_dict

    def _on_table_selection_changed(self):
        """Update layer selection when table cells are selected."""
        if self._selection_blocked:
            return

        # Get all selected row indices from the table
        selected_global_rows = [
            self.table.model().mapToSource(raw_index).row()
            for raw_index in self.table.selectionModel().selectedIndexes()
        ]

        if not selected_global_rows:
            return

        df = self.table.model().sourceModel().df
        if df.empty:
            return

        # Handle single layer case (no 'Layer' column)
        if len(self._selected_layers) == 1:
            layer = self._selected_layers[0]
            # All rows belong to this layer; global row indices are layer-specific indices
            layer_specific_indices = np.array(selected_global_rows)

            with self._block_selection():
                if hasattr(layer, 'selected_label'):
                    layer.selected_label = layer_specific_indices[-1]
                elif hasattr(layer, 'selected_data'):
                    layer.selected_data = set(layer_specific_indices)
            return

        # Handle multiple layers case (has source column)
        # Indices may not be continuous, so use positional indexing
        selections_by_layer = {}
        source_col_name = df.columns[
            0
        ]  # Source column is always first when present

        for row_pos in selected_global_rows:
            # Get source identifier for this row
            layer_name = df.iloc[row_pos][source_col_name]
            # Convert global row to layer-specific row
            layer_row_idx = self._global_row_to_layer_row(
                df, row_pos, layer_name
            )
            selections_by_layer.setdefault(layer_name, []).append(
                layer_row_idx
            )

        # Update layer selections
        for layer in self._selected_layers:
            if layer.name not in selections_by_layer:
                continue
            layer_indices = selections_by_layer[layer.name]

            with self._block_selection():
                if hasattr(layer, 'selected_label'):
                    layer.selected_label = layer_indices[-1]
                elif hasattr(layer, 'selected_data'):
                    layer.selected_data = set(layer_indices)

    def _update_table_selected_cells(self):
        """Update table selected cells when layer selection changes."""
        if self._selection_blocked:
            return

        model = self.table.model()
        df = model.sourceModel().df
        if df.empty:
            return

        indices = []

        # Handle single layer case (no 'Layer' column)
        if len(self._selected_layers) == 1:
            layer = self._selected_layers[0]
            # All rows belong to this single layer
            # layer.selected_data/selected_label refer to positions in the layer

            if hasattr(layer, 'selected_label'):
                sel = layer.selected_label
                # Position in layer maps directly to df position
                indices += [[sel]]
            elif hasattr(layer, 'selected_data'):
                sel = layer.selected_data
                # Positions in layer map directly to df positions
                indices += [list(sel)]
        else:
            # Handle multiple sources case
            for layer in self._selected_layers:
                if hasattr(layer, 'selected_label'):
                    sel = layer.selected_label
                    # Position within the layer's features
                    global_row = self._layer_row_to_global_row(
                        df, sel, layer.name
                    )
                    if global_row is not None:
                        indices += [[global_row]]
                elif hasattr(layer, 'selected_data'):
                    sel = layer.selected_data
                    # Positions within the layer's features
                    for layer_pos in sel:
                        global_row = self._layer_row_to_global_row(
                            df, layer_pos, layer.name
                        )
                        if global_row is not None:
                            indices += [[global_row]]
                else:
                    continue

        if not indices:
            # deselect all rows if no selection attributes are found
            self.table.selectionModel().clearSelection()
            self.table.viewport().update()
            return  # nothing to select

        indices = np.concatenate(indices)
        selection = QItemSelection()

        for idx in indices:
            idx = model.mapFromSource(model.sourceModel().index(idx, 0))
            # select whole rows
            first_cell = model.index(idx.row(), 0)
            last_cell = model.index(idx.row(), model.columnCount() - 1)
            selection.select(first_cell, last_cell)

        # block signals while setting to prevent infinite loop
        with self._block_selection():
            self.table.selectionModel().select(
                selection,
                QItemSelectionModel.SelectionFlag.ClearAndSelect
                | QItemSelectionModel.SelectionFlag.Rows,
            )

        self.table.viewport().update()

    def _on_table_data_changed(self, topLeft, bottomRight, roles=None):
        """Sync edits in the table back to the correct layer.features DataFrame."""
        model = self.table.model().sourceModel()
        df = model.df
        # For each edited cell
        for row in range(topLeft.row(), bottomRight.row() + 1):
            # Determine which layer this row belongs to
            if len(self._selected_layers) == 1:
                # Single layer case: all rows belong to this layer
                layer = self._selected_layers[0]
                # Row position directly maps to layer position
                layer_row_idx = row
            else:
                # Multiple sources case: find source from the source column
                source_col_name = df.columns[
                    0
                ]  # Source column is always first when present
                layer_name = df.iloc[row][source_col_name]
                layer = next(
                    ly for ly in self._selected_layers if ly.name == layer_name
                )
                # Convert global row to layer-specific row
                layer_row_idx = self._global_row_to_layer_row(
                    df, row, layer_name
                )

            # Update the layer features DataFrame (except if immutable columns)
            for col in range(topLeft.column(), bottomRight.column() + 1):
                if model.is_column_immutable(col):
                    continue
                # Map view column to DataFrame column
                df_col_idx = model._get_df_col_idx(col)
                if df_col_idx is None:
                    continue
                col_name = df.columns[df_col_idx]
                if col_name not in layer.features.columns:
                    continue  # do not update features if column not present
                layer.features.iloc[
                    layer_row_idx, layer.features.columns.get_loc(col_name)
                ] = df.iloc[row, df_col_idx]

    @contextmanager
    def _block_selection(self):
        self._selection_blocked = True
        yield
        self._selection_blocked = False

    def _global_row_to_layer_row(
        self, df: pd.DataFrame, global_row: int, layer_name: str
    ) -> int:
        """Convert global DataFrame row position to layer-specific row position.

        Parameters
        ----------
        df : pd.DataFrame
            The combined features DataFrame
        global_row : int
            Row position in the combined DataFrame
        layer_name : str
            Name of the layer

        Returns
        -------
        int
            Row position within the layer's features
        """
        source_col_name = df.columns[0]
        matching_rows = df[df[source_col_name] == layer_name]
        layer_row_positions = matching_rows.index.tolist()
        df_row_index_value = df.index[global_row]

        # Find all occurrences of this index value in the layer
        layer_local_positions = [
            i
            for i, idx_val in enumerate(layer_row_positions)
            if idx_val == df_row_index_value
        ]

        # Handle duplicate indices by checking actual DataFrame position
        if len(layer_local_positions) == 1:
            return layer_local_positions[0]
        # Multiple rows with same index value in this layer
        layer_start_pos = df.index.tolist().index(layer_row_positions[0])
        return global_row - layer_start_pos

    def _layer_row_to_global_row(
        self, df: pd.DataFrame, layer_row: int, layer_name: str
    ) -> int | None:
        """Convert layer-specific row position to global DataFrame row position.

        Parameters
        ----------
        df : pd.DataFrame
            The combined features DataFrame
        layer_row : int
            Row position within the layer's features
        layer_name : str
            Name of the layer

        Returns
        -------
        int | None
            Row position in the combined DataFrame, or None if not found
        """
        source_col_name = df.columns[0]
        matching_rows = df[df[source_col_name] == layer_name]

        if layer_row >= len(matching_rows):
            return None

        # Get the index value at this layer position
        global_pos = matching_rows.iloc[[layer_row]].index.tolist()[0]

        # Find this index value's position in the full DataFrame
        df_positions = df.index.tolist()
        layer_positions = [
            i
            for i, idx in enumerate(df_positions)
            if idx == global_pos and df.iloc[i][source_col_name] == layer_name
        ]

        return layer_positions[0] if layer_positions else None

    def _on_save_clicked(self):
        dlg = QFileDialog()
        hist = get_save_history()
        dlg.setHistory(hist)

        # Generate filename for multiple layers
        if len(self._selected_layers) == 1:
            fname = f'{self._selected_layers[0].name}_features.csv'
        elif len(self._selected_layers) > 1:
            first_layer = self._selected_layers[0].name
            others_count = len(self._selected_layers) - 1
            fname = (
                f'{first_layer}_and_{others_count}_other_layers_features.csv'
            )

        fname = self._remove_invalid_chars(fname)

        fname, _ = dlg.getSaveFileName(
            self,  # parent
            'Save layer features',  # caption
            str(Path(hist[0]) / fname),  # directory in PyQt, dir in PySide
            filter='*.csv',
            options=(
                QFileDialog.Option.DontUseNativeDialog
                if in_ipython()
                else QFileDialog.Options()
            ),
        )

        if not fname:
            # was closed without saving
            return  # pragma: no cover

        df = self.table.model().sourceModel().df
        df.to_csv(fname)

    # copied from QtViewer
    def _remove_invalid_chars(self, selected_layer_name):
        """Removes invalid characters from selected layer name to suggest a filename.

        Parameters
        ----------
        selected_layer_name : str
            The selected napari layer name.

        Returns
        -------
        suggested_name : str
            Suggested name from input selected layer name, without invalid characters.
        """
        unprintable_ascii_chars = (
            '\x00',
            '\x01',
            '\x02',
            '\x03',
            '\x04',
            '\x05',
            '\x06',
            '\x07',
            '\x08',
            '\x0e',
            '\x0f',
            '\x10',
            '\x11',
            '\x12',
            '\x13',
            '\x14',
            '\x15',
            '\x16',
            '\x17',
            '\x18',
            '\x19',
            '\x1a',
            '\x1b',
            '\x1c',
            '\x1d',
            '\x1e',
            '\x1f',
            '\x7f',
        )
        invalid_characters = (
            ''.join(unprintable_ascii_chars)
            + '/'
            + '\\'  # invalid Windows filename character
            + ':*?"<>|\t\n\r\x0b\x0c'  # invalid Windows path characters
        )
        translation_table = dict.fromkeys(map(ord, invalid_characters), None)
        # Remove invalid characters
        suggested_name = selected_layer_name.translate(translation_table)
        return suggested_name



"""
magicgui Image Arithmetic
=========================

Basic example of using magicgui to create an Image Arithmetic GUI in napari.

.. tags:: gui
"""

import enum

import numpy as np

import napari


# Enums are a convenient way to get a dropdown menu
class Operation(enum.Enum):
    """A set of valid arithmetic operations for image_arithmetic."""

    add = np.add
    subtract = np.subtract
    multiply = np.multiply
    divide = np.divide


# Define our image_arithmetic function.
# Note that we can use forward references for the napari type annotations.
# You can read more about them here:
# https://peps.python.org/pep-0484/#forward-references
# In this example, because we have already imported napari anyway, it doesn't
# really matter. But this syntax would let you specify that a parameter is a
# napari object type without actually importing or depending on napari.
# Note: here we use `napari.types.ImageData` as our parameter annotations,
# which means our function will be passed layer.data instead of
# the full layer instance
def image_arithmetic(
    layerA: 'napari.types.ImageData',
    operation: Operation,
    layerB: 'napari.types.ImageData',
) -> 'napari.types.ImageData':
    """Adds, subtracts, multiplies, or divides two same-shaped image layers."""
    if layerA is not None and layerB is not None:
        return operation.value(layerA, layerB)
    return None


# create a new viewer with a couple image layers
# Since the input layers have been scaled, the resulting returned layer will inherit the same scale value.
viewer = napari.Viewer()
viewer.add_image(np.random.rand(20, 20), name='Layer 1', scale=(2, 2))
viewer.add_image(np.random.rand(20, 20), name='Layer 2', scale=(2, 2))

# Add our magic function to napari
viewer.window.add_function_widget(image_arithmetic)

if __name__ == '__main__':
    napari.run()



"""
Add points with features
========================

Display a points layer on top of an image layer using the ``add_points`` and
``add_image`` APIs

.. tags:: visualization-basic, features-table
"""

import numpy as np
from skimage import data
from skimage.color import rgb2gray

import napari

# add the image
viewer = napari.Viewer()
layer = viewer.add_image(rgb2gray(data.astronaut()))
# add the points
points = np.array([[100, 100], [200, 200], [333, 111]])

# create features for each point
features = {
    'confidence': np.array([1, 0.5, 0]),
    'good_point': np.array([True, False, False])
}

# define the color cycle for the face_color annotation
face_color_cycle = ['blue', 'green']

# create a points layer where the face_color is set by the good_point feature
# and the border_color is set via a color map (grayscale) on the confidence
# feature.
points_layer = viewer.add_points(
    points,
    features=features,
    size=20,
    border_width=7,
    border_width_is_relative=False,
    border_color='confidence',
    border_colormap='gray',
    face_color='good_point',
    face_color_cycle=face_color_cycle
)

# set the border_color mode to colormap
points_layer.border_color_mode = 'colormap'


# bind a function to toggle the good_point annotation of the selected points
@viewer.bind_key('t')
def toggle_point_annotation(viewer):
    selected_points = list(points_layer.selected_data)
    if len(selected_points) > 0:
        good_point = points_layer.features['good_point']
        good_point[selected_points] = ~good_point[selected_points]
        points_layer.features['good_point'] = good_point

        # we need to manually refresh since we did not use the Points.features
        # setter to avoid changing the color map if all points get toggled to
        # the same class, we set update_colors=False (only re-colors the point
        # using the previously-determined color mapping).
        points_layer.refresh_colors(update_color_mapping=False)


if __name__ == '__main__':
    napari.run()



"""
Comparison of Screenshot and Figure Export
==========================================

Display multiple layer types, add scale bar, and take a screenshot or export a
figure from a 'light' canvas. Then switch to a 'dark' canvas and display the
screenshot and figure. Compare the limits of each export method. The screenshot
will include the entire canvas, and results in some layers being clipped
if it extends outside the canvas. This also means that screenshots will
reflect the current zoom. In comparison, the `export_figure` will always
include the extent of the layers and any other elements overlayed
on the canvas, such as the scale bar. Exported figures also move the scale bar
to within the margins of the canvas.

In the final grid state shown below, the first row represents exported images. The first two show that zoom is not reflected in the exported figure. The final one shows how the exported figure adapts to change in the layer extent. In the second row are the screenshots, showing the fact that the entire canvas is captured and that zoom is preserved.

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data

import napari

# Create a napari viewer with multiple layer types and add a scale bar.
# One of the polygon shapes exists outside the image extent, which is
# useful in displaying how figure export handles the extent of all layers.

viewer = napari.Viewer()

# add a 2D image layer
img_layer = viewer.add_image(data.camera(), name='photographer')
img_layer.colormap = 'gray'

# polygon within image extent
layer_within = viewer.add_shapes(
    np.array([[11, 13], [111, 113], [22, 246]]),
    shape_type='polygon',
    face_color='coral',
    name='shapes_within',
)

# add a polygon shape layer
layer_outside = viewer.add_shapes(
    np.array([[572, 222], [305, 292], [577, 440]]),
    shape_type='polygon',
    face_color='royalblue',
    name='shapes_outside',
)

# add scale_bar with background box
viewer.scale_bar.visible = True
viewer.scale_bar.box = True
# viewer.scale_bar.length = 150  # prevent dynamic adjustment of scale bar length


# Take screenshots and export figures in 'light' theme, to show the canvas
# margins and the extent of the exported figure.
viewer.theme = 'light'
screenshot = viewer.screenshot()
figure = viewer.export_figure()
# optionally, save the exported figure: viewer.export_figure(path='export_figure.png')
# or screenshot: viewer.screenshot(path='screenshot.png')


# Zoom in and take another screenshot and export figure to show the different
# extents of the exported figure and screenshot.
viewer.camera.zoom = 3
screenshot_zoomed = viewer.screenshot()
figure_zoomed = viewer.export_figure()


# Remove the layer that exists outside the image extent and take another
# figure export to show the extent of the exported figure without the
# layer that exists outside the camera image extent.
viewer.layers.remove(layer_outside)
figure_no_outside_shape = viewer.export_figure()


# Display the screenshots and figures in 'dark' theme, and switch to grid mode
# for comparison. In the final grid state shown, the first row represents exported
# images. The first two show that zoom is not reflected in the exported figure.
# The final one shows how the exported figure adapts to change in the layer extent.
# In the second row are the screenshots, showing the fact that the entire canvas
# is captured and that zoom is preserved.
viewer.theme = 'dark'
viewer.layers.select_all()
viewer.layers.remove_selected()

viewer.add_image(screenshot_zoomed, rgb=True, name='screenshot_zoomed')
viewer.add_image(screenshot, rgb=True, name='screenshot')
viewer.add_image(figure_no_outside_shape, rgb=True, name='figure_no_outside_shape')
viewer.add_image(figure_zoomed, rgb=True, name='figure_zoomed')
viewer.add_image(figure, rgb=True, name='figure')

viewer.grid.enabled = True
viewer.grid.shape = (2, 3)

if __name__ == '__main__':
    napari.run()



"""
Tracks 3D
=========

.. tags:: visualization-advanced
"""

import numpy as np

import napari


def lissajous(t):
    a = np.random.random(size=(3,)) * 80.0 - 40.0
    b = np.random.random(size=(3,)) * 0.05
    c = np.random.random(size=(3,)) * 0.1
    return (a[i] * np.cos(b[i] * t + c[i]) for i in range(3))


def tracks_3d(num_tracks=10):
    """ create 3d+t track data """
    tracks = []

    for track_id in range(num_tracks):

        # space to store the track data and features
        track = np.zeros((200, 10), dtype=np.float32)

        # time
        timestamps = np.arange(track.shape[0])
        x, y, z = lissajous(timestamps)

        track[:, 0] = track_id
        track[:, 1] = timestamps
        track[:, 2] = 50.0 + z
        track[:, 3] = 50.0 + y
        track[:, 4] = 50.0 + x

        # calculate the speed as a feature
        gz = np.gradient(track[:, 2])
        gy = np.gradient(track[:, 3])
        gx = np.gradient(track[:, 4])

        speed = np.sqrt(gx ** 2 + gy ** 2 + gz ** 2)
        distance = np.sqrt(x ** 2 + y ** 2 + z ** 2)

        track[:, 5] = gz
        track[:, 6] = gy
        track[:, 7] = gx
        track[:, 8] = speed
        track[:, 9] = distance

        tracks.append(track)

    tracks = np.concatenate(tracks, axis=0)
    data = tracks[:, :5]  # just the coordinate data

    features = {
        'time': tracks[:, 1],
        'gradient_z': tracks[:, 5],
        'gradient_y': tracks[:, 6],
        'gradient_x': tracks[:, 7],
        'speed': tracks[:, 8],
        'distance': tracks[:, 9],
    }

    graph = {}
    return data, features, graph


tracks, features, graph = tracks_3d(num_tracks=100)
vertices = tracks[:, 1:]

viewer = napari.Viewer(ndisplay=3)
viewer.add_points(vertices, size=1, name='points', opacity=0.3)
viewer.add_tracks(tracks, features=features, name='tracks')

if __name__ == '__main__':
    napari.run()



import dask.array as da
import zarr

import napari

"""
The sample data here is Interphase HeLa Cell [https://openorganelle.janelia.org/datasets/jrc_hela-3],
from HHMI's OpenOrganelle [https://openorganelle.janelia.org].

The data are hosted by Open Data on AWS on S3.

This tests access to multi-scale remote data.
"""


# access the root of the n5 container
group = zarr.open(zarr.N5FSStore('s3://janelia-cosem-datasets/jrc_hela-2/jrc_hela-2.n5', anon=True))

# s0 (highest resolution) through s5 (lowest resolution) are available
data = []
for i in range(5):
    zarr_array = group[f'em/fibsem-uint16/s{i}']
    data.append(da.from_zarr(zarr_array, chunks=zarr_array.chunks))

# This order presents a better visualization, but seems to break simple async (issue #5106)
# layer = viewer.add_image(data, order=(1, 0, 2), contrast_limits=(18000, 40000), multiscale=True)
viewer = napari.Viewer()
layer = viewer.add_image(data, contrast_limits=(18000, 40000), multiscale=True)

if __name__ == '__main__':
    napari.run()



from napari._qt.layer_controls.widgets._tracks.qt_color_properties_combobox import (
    QtColorPropertiesComboBoxControl,
)
from napari._qt.layer_controls.widgets._tracks.qt_colormap_control import (
    QtColormapComboBoxControl,
)
from napari._qt.layer_controls.widgets._tracks.qt_graph_checkbox import (
    QtGraphCheckBoxControl,
)
from napari._qt.layer_controls.widgets._tracks.qt_head_slider import (
    QtHeadLengthSliderControl,
)
from napari._qt.layer_controls.widgets._tracks.qt_hide_completed_tracks_checkbox import (
    QtHideCompletedTracksCheckBoxControl,
)
from napari._qt.layer_controls.widgets._tracks.qt_id_checkbox import (
    QtIdCheckBoxControl,
)
from napari._qt.layer_controls.widgets._tracks.qt_tail_control import (
    QtTailDisplayCheckBoxControl,
    QtTailLengthSliderControl,
    QtTailWidthSliderControl,
)

__all__ = [
    'QtColorPropertiesComboBoxControl',
    'QtColormapComboBoxControl',
    'QtGraphCheckBoxControl',
    'QtHeadLengthSliderControl',
    'QtHideCompletedTracksCheckBoxControl',
    'QtIdCheckBoxControl',
    'QtTailDisplayCheckBoxControl',
    'QtTailLengthSliderControl',
    'QtTailWidthSliderControl',
]



"""
Zarr array
==========

Display a zarr array

.. tags:: visualization-nD
"""

try:
    import zarr
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        """This example uses a zarr array but zarr is not
    installed. To install try 'pip install zarr'."""
    ) from None

import napari

data = zarr.zeros((102_0, 200, 210), chunks=(100, 200, 210))
data[53_0:53_1, 100:110, 110:120] = 1

print(data.shape)
# For big data, we should specify the contrast_limits range, or napari will try
# to find the min and max of the full image.
viewer = napari.Viewer()
layer = viewer.add_image(data, contrast_limits=[0, 1], rgb=False)

if __name__ == '__main__':
    napari.run()



"""
New theme
=========

Displays an image and sets the theme to new custom theme.

.. tags:: experimental
"""

from skimage import data

import napari
from napari.utils.theme import available_themes, get_theme, register_theme

# create the viewer with an image
viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True, name='astronaut')

# List themes
print('Originally themes', available_themes())

blue_theme = get_theme('dark')
blue_theme.id = 'blue'
blue_theme.icon = (
    'rgb(0, 255, 255)'  # you can provide colors as rgb(XXX, YYY, ZZZ)
)
blue_theme.background = 28, 31, 48  # or as tuples
blue_theme.foreground = [45, 52, 71]  # or as list
blue_theme.primary = '#50586c'  # or as hexes
blue_theme.current = 'orange'  # or as color name
blue_theme.font_size = '10pt'  # you can provide a font size in points (pt) for the application

register_theme('blue', blue_theme, 'custom')

# List themes
print('New themes', available_themes())

# Set theme
viewer.theme = 'blue'

if __name__ == '__main__':
    napari.run()



import numpy as np

from napari._vispy.layers.base import VispyBaseLayer
from napari._vispy.utils.gl import BLENDING_MODES
from napari._vispy.utils.text import update_text
from napari._vispy.visuals.points import PointsVisual
from napari.settings import get_settings
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.events import disconnect_events


class VispyPointsLayer(VispyBaseLayer):
    node: PointsVisual

    def __init__(self, layer) -> None:
        node = PointsVisual()
        super().__init__(layer, node)

        self.layer.events.symbol.connect(self._on_data_change)
        self.layer.events.border_width.connect(self._on_data_change)
        self.layer.events.border_width_is_relative.connect(
            self._on_data_change
        )
        self.layer.events.border_color.connect(self._on_data_change)
        self.layer._border.events.colors.connect(self._on_data_change)
        self.layer._border.events.color_properties.connect(
            self._on_data_change
        )
        self.layer.events.face_color.connect(self._on_data_change)
        self.layer._face.events.colors.connect(self._on_data_change)
        self.layer._face.events.color_properties.connect(self._on_data_change)
        self.layer.events.highlight.connect(self._on_highlight_change)
        self.layer.text.events.connect(self._on_text_change)
        self.layer.events.shading.connect(self._on_shading_change)
        self.layer.events.antialiasing.connect(self._on_antialiasing_change)
        self.layer.events.canvas_size_limits.connect(
            self._on_canvas_size_limits_change
        )
        self.layer.events.scale_factor.connect(self._update_text)

        self._on_data_change()

    def _on_data_change(self):
        # Set vispy data, noting that the order of the points needs to be
        # reversed to make the most recently added point appear on top
        # and the rows / columns need to be switched for vispy's x / y ordering
        if len(self.layer._indices_view) == 0:
            # always pass one invisible point to avoid issues
            data = np.zeros((1, self.layer._slice_input.ndisplay))
            size = np.zeros(1)
            border_color = np.array([[0.0, 0.0, 0.0, 1.0]], dtype=np.float32)
            face_color = np.array([[1.0, 1.0, 1.0, 1.0]], dtype=np.float32)
            border_width = np.zeros(1)
            symbol = ['o']
        else:
            data = self.layer._view_data
            size = self.layer._view_size
            border_color = self.layer._view_border_color
            face_color = self.layer._view_face_color
            border_width = self.layer._view_border_width
            symbol = [str(x) for x in self.layer._view_symbol]

        set_data = self.node.points_markers.set_data

        # use only last dimension to scale point sizes, see #5582
        scale = self.layer.scale[-1]
        scaled_size = size * scale

        if self.layer.border_width_is_relative:
            border_kw = {
                'edge_width': None,
                'edge_width_rel': border_width,
            }
        else:
            # cap border_width to the marker size, cause otherwise we get strange effects
            # in the shaders which result in unbound growth of highlights)
            border_kw = {
                'edge_width': np.minimum(border_width * scale, scaled_size),
                'edge_width_rel': None,
            }

        set_data(
            data[:, ::-1],
            size=size * scale,
            symbol=symbol,
            # edge_color is the name of the vispy marker visual kwarg
            edge_color=border_color,
            face_color=face_color,
            **border_kw,
        )

        self.reset()

    def _on_highlight_change(self):
        settings = get_settings()
        if len(self.layer._highlight_index) > 0:
            # Color the hovered or selected points

            # _highlight_index contains indices into the view arrays, but we can get the
            # actual data indices once to avoid materializing the entire view for each property
            data_indices = self.layer._indices_view[
                self.layer._highlight_index
            ]

            data = self.layer.data[
                np.ix_(data_indices, self.layer._slice_input.displayed)
            ]
            if data.ndim == 1:
                data = np.expand_dims(data, axis=0)
            size = self.layer.size[data_indices] * self.layer._view_size_scale
            border_width = self.layer.border_width[data_indices]
            if self.layer.border_width_is_relative:
                border_width = border_width * size
            symbol = self.layer.symbol[data_indices]
        else:
            data = np.empty((0, self.layer._slice_input.ndisplay))
            size = 0
            symbol = ['o']
            border_width = np.empty(0)

        scale = self.layer.scale[-1]
        highlight_thickness = settings.appearance.highlight.highlight_thickness
        scaled_size = (size + border_width) * scale
        # cap scaled_highlight to the marker size, cause otherwise we get strange effects
        # in the shaders which result in unbound growth of highlights)
        scaled_highlight = np.minimum(
            highlight_thickness * self.layer.scale_factor, scaled_size
        )
        highlight_color = tuple(settings.appearance.highlight.highlight_color)

        self.node.selection_markers.set_data(
            data[:, ::-1],
            size=scaled_size,
            symbol=symbol,
            edge_width=scaled_highlight * 2,
            edge_color=highlight_color,
            face_color=transform_color('transparent'),
        )

        if (
            self.layer._highlight_box is None
            or 0 in self.layer._highlight_box.shape
        ):
            pos = np.zeros((1, 3))
            highlight_thickness = 0
        else:
            pos = self.layer._highlight_box
            # add z=0 if 2d (see #6819)
            if pos.shape[1] == 2:
                pos = np.pad(pos, ((0, 0), (1, 0)))

        self.node.highlight_lines.set_data(
            pos=pos[:, ::-1],
            color=highlight_color,
            width=highlight_thickness,
        )

        self.node.update()

    def _update_text(self, *, update_node=True):
        """Function to update the text node properties

        Parameters
        ----------
        update_node : bool
            If true, update the node after setting the properties
        """
        update_text(node=self.node.text, layer=self.layer)
        if update_node:
            self.node.update()

    def _on_text_change(self, event=None):
        if event is not None:
            if event.type == 'blending':
                self._on_blending_change(event)
                return
            if event.type == 'values':
                return
        self._update_text()

    def _on_blending_change(self, event=None):
        """Function to set the blending mode"""
        super()._on_blending_change()

        text_blending_kwargs = BLENDING_MODES[self.layer.text.blending]
        self.node.text.set_gl_state(**text_blending_kwargs)

        # selection box is always without depth
        box_blending_kwargs = BLENDING_MODES['translucent_no_depth']
        self.node.highlight_lines.set_gl_state(**box_blending_kwargs)

        self.node.update()

    def _on_antialiasing_change(self):
        self.node.antialias = self.layer.antialiasing

    def _on_shading_change(self):
        shading = self.layer.shading
        self.node.spherical = shading == 'spherical'

    def _on_canvas_size_limits_change(self):
        self.node.points_markers.canvas_size_limits = (
            self.layer.canvas_size_limits
        )
        highlight_thickness = (
            get_settings().appearance.highlight.highlight_thickness
        )
        low, high = self.layer.canvas_size_limits
        self.node.selection_markers.canvas_size_limits = (
            low + highlight_thickness,
            high + highlight_thickness,
        )
        self.node.update()

    def reset(self):
        super().reset()
        self._update_text(update_node=False)
        self._on_highlight_change()
        self._on_antialiasing_change()
        self._on_shading_change()
        self._on_canvas_size_limits_change()

    def close(self):
        """Vispy visual is closing."""
        disconnect_events(self.layer.text.events, self)
        super().close()



import threading
import warnings
from concurrent.futures import Future
from dataclasses import dataclass
from unittest.mock import MagicMock

import dask.array as da
import pytest
from qtpy.QtCore import Qt, QThread
from qtpy.QtWidgets import QPushButton, QWidget

from napari._qt.dialogs.qt_notification import (
    NapariQtNotification,
    TracebackDialog,
)
from napari._tests.utils import DEFAULT_TIMEOUT_SECS, skip_on_win_ci
from napari.utils.notifications import (
    ErrorNotification,
    Notification,
    NotificationSeverity,
    notification_manager,
)


def _threading_warn():
    thr = threading.Thread(target=_warn)
    thr.start()
    thr.join(timeout=DEFAULT_TIMEOUT_SECS)


def _warn():
    warnings.warn('warning!', stacklevel=3)


def _threading_raise():
    thr = threading.Thread(target=_raise)
    thr.start()
    thr.join(timeout=DEFAULT_TIMEOUT_SECS)


def _raise():
    raise ValueError('error!')


@pytest.fixture
def _clean_current(monkeypatch, qtbot):
    from napari._qt.qt_main_window import _QtMainWindow

    base_show = NapariQtNotification.show

    widget = QWidget()
    qtbot.addWidget(widget)
    mock_window = MagicMock()
    widget.resized = MagicMock()
    mock_window._qt_viewer.canvas = widget
    mock_window._qt_viewer.canvas.native = widget

    def mock_current_main_window(*_, **__):
        """
        This return mock main window object to ensure that
        notification dialog has parent added to qtbot
        """
        return mock_window

    def store_widget(self, *args, **kwargs):
        base_show(self, *args, **kwargs)

    monkeypatch.setattr(NapariQtNotification, 'show', store_widget)
    monkeypatch.setattr(_QtMainWindow, 'current', mock_current_main_window)


@dataclass
class ShowStatus:
    show_notification_count: int = 0
    show_traceback_count: int = 0


@pytest.fixture(autouse=True)
def _raise_on_show(monkeypatch, qtbot):
    def raise_prepare(text):
        def _raise_on_call(self, *args, **kwargs):
            raise RuntimeError(text)

        return _raise_on_call

    monkeypatch.setattr(
        NapariQtNotification, 'show', raise_prepare('notification show')
    )
    monkeypatch.setattr(
        TracebackDialog, 'show', raise_prepare('traceback show')
    )
    monkeypatch.setattr(
        NapariQtNotification,
        'close_with_fade',
        raise_prepare('close_with_fade'),
    )


@pytest.fixture
def count_show(monkeypatch, qtbot):
    stat = ShowStatus()

    def mock_show_notif(_):
        stat.show_notification_count += 1

    def mock_show_traceback(_):
        stat.show_traceback_count += 1

    monkeypatch.setattr(NapariQtNotification, 'show', mock_show_notif)
    monkeypatch.setattr(TracebackDialog, 'show', mock_show_traceback)

    return stat


@pytest.fixture(autouse=True)
def _ensure_qtbot(monkeypatch, qtbot):
    old_notif_init = NapariQtNotification.__init__
    old_traceback_init = TracebackDialog.__init__

    def mock_notif_init(self, *args, **kwargs):
        old_notif_init(self, *args, **kwargs)
        qtbot.add_widget(self)

    def mock_traceback_init(self, *args, **kwargs):
        old_traceback_init(self, *args, **kwargs)
        qtbot.add_widget(self)

    monkeypatch.setattr(NapariQtNotification, '__init__', mock_notif_init)
    monkeypatch.setattr(TracebackDialog, '__init__', mock_traceback_init)


@pytest.mark.skip('debug')
def test_clean_current_path_exist(make_napari_viewer):
    """If this test fail then you need to fix also clean_current fixture"""
    assert isinstance(make_napari_viewer().window._qt_viewer, QWidget)


@pytest.mark.parametrize(
    ('raise_func', 'warn_func'),
    [(_raise, _warn), (_threading_raise, _threading_warn)],
)
def test_notification_manager_via_gui(
    count_show, qtbot, raise_func, warn_func, monkeypatch
):
    """
    Test that the notification_manager intercepts `sys.excepthook`` and
    `threading.excepthook`.
    """
    errButton = QPushButton()
    warnButton = QPushButton()
    errButton.clicked.connect(raise_func)
    warnButton.clicked.connect(warn_func)
    qtbot.addWidget(errButton)
    qtbot.addWidget(warnButton)
    monkeypatch.setattr(
        NapariQtNotification, 'show_notification', lambda x: None
    )
    with notification_manager:
        for btt, expected_message in [
            (errButton, 'error!'),
            (warnButton, 'warning!'),
        ]:
            notification_manager.records = []
            qtbot.mouseClick(btt, Qt.MouseButton.LeftButton)
            assert len(notification_manager.records) == 1
            assert notification_manager.records[0].message == expected_message
            notification_manager.records = []


@pytest.mark.usefixtures('_clean_current')
def test_show_notification_from_thread(count_show, monkeypatch, qtbot):
    from napari.settings import get_settings

    settings = get_settings()

    monkeypatch.setattr(
        settings.application,
        'gui_notification_level',
        NotificationSeverity.INFO,
    )

    class CustomThread(QThread):
        def run(self):
            notif = Notification(
                'hi',
                NotificationSeverity.INFO,
                actions=[('click', lambda x: None)],
            )
            res = NapariQtNotification.show_notification(notif)
            assert isinstance(res, Future)
            assert res.result(timeout=DEFAULT_TIMEOUT_SECS) is None
            assert count_show.show_notification_count == 1

    thread = CustomThread()
    with qtbot.waitSignal(thread.finished):
        thread.start()


@pytest.mark.usefixtures('_clean_current')
@pytest.mark.parametrize('severity', NotificationSeverity.__members__)
def test_notification_display(count_show, severity, monkeypatch):
    """Test that NapariQtNotification can present a Notification event.

    NOTE: in napari.utils._tests.test_notification_manager, we already test
    that the notification manager successfully overrides sys.excepthook,
    and warnings.showwarning... and that it emits an event which is an instance
    of napari.utils.notifications.Notification.

    in `get_qapp()`, we connect `notification_manager.notification_ready` to
    `NapariQtNotification.show_notification`, so all we have to test here is
    that show_notification is capable of receiving various event types.
    (we don't need to test that )
    """
    from napari.settings import get_settings

    settings = get_settings()

    monkeypatch.delenv('NAPARI_CATCH_ERRORS', raising=False)
    monkeypatch.setattr(
        settings.application,
        'gui_notification_level',
        NotificationSeverity.INFO,
    )
    notif = Notification('hi', severity, actions=[('click', lambda x: None)])
    NapariQtNotification.show_notification(notif)
    if NotificationSeverity(severity) >= NotificationSeverity.INFO:
        assert count_show.show_notification_count == 1
    else:
        assert count_show.show_notification_count == 0

    dialog = NapariQtNotification.from_notification(notif)
    assert not dialog.property('expanded')
    dialog.toggle_expansion()
    assert dialog.property('expanded')
    dialog.toggle_expansion()
    assert not dialog.property('expanded')


def test_notification_error(count_show, monkeypatch):
    from napari.settings import get_settings

    settings = get_settings()

    monkeypatch.delenv('NAPARI_CATCH_ERRORS', raising=False)
    monkeypatch.setattr(
        NapariQtNotification, 'close_with_fade', lambda x, y: None
    )
    monkeypatch.setattr(
        settings.application,
        'gui_notification_level',
        NotificationSeverity.INFO,
    )
    try:
        raise ValueError('error!')
    except ValueError as e:
        notif = ErrorNotification(e)

    dialog = NapariQtNotification.from_notification(notif)

    bttn = dialog.row2_widget.findChild(QPushButton)
    assert bttn.text() == 'View Traceback'
    assert count_show.show_traceback_count == 0
    bttn.click()
    assert count_show.show_traceback_count == 1


@skip_on_win_ci
@pytest.mark.usefixtures('_clean_current')
def test_notifications_error_with_threading(make_napari_viewer, monkeypatch):
    """Test notifications of `threading` threads, using a dask example."""
    random_image = da.random.random((10, 10))
    monkeypatch.setattr(
        NapariQtNotification, 'show_notification', lambda x: None
    )
    with notification_manager:
        viewer = make_napari_viewer(strict_qt=False)
        viewer.add_image(random_image)
        result = da.divide(random_image, da.zeros((10, 10)))
        viewer.add_image(result)
        assert len(notification_manager.records) >= 1
        notification_manager.records = []



"""
Add points on nD shapes
=======================

Add points on nD shapes in 3D using a mouse callback

.. tags:: visualization-nD
"""

import numpy as np

import napari

# Create rectangles in 4D
data = [
        [
            [0, 50, 75, 75],
            [0, 50, 125, 75],
            [0, 100, 125, 125],
            [0, 100, 75, 125]
        ],
        [
            [0, 10, 75, 75],
            [0, 10, 125, 75],
            [0, 40, 125, 125],
            [0, 40, 75, 125]
        ],
        [
            [1, 100, 75, 75],
            [1, 100, 125, 75],
            [1, 50, 125, 125],
            [1, 50, 75, 125]
        ]
    ]

shapes_data = np.array(data)

# add an empty 4d points layer
viewer = napari.Viewer()
points_layer = viewer.add_points(ndim=4, size=3)
# add the shapes layer to the viewer
features = {'index': [0, 1, 2]}
for shape_type, mult in {('ellipse', 1), ('rectangle', -1)}:
    shapes_layer = viewer.add_shapes(
        shapes_data * mult,
        face_color=['magenta', 'green', 'blue'],
        edge_color='white',
        blending='additive',
        features=features,
        text='index',
        shape_type=shape_type,
    )

    @shapes_layer.mouse_drag_callbacks.append
    def on_click(layer, event):

        shape_index, intersection_point = layer.get_index_and_intersection(
            event.position, event.view_direction, event.dims_displayed
        )

        if (shape_index is not None) and (intersection_point is not None):
            points_layer.add(intersection_point)


for d in data:
    viewer.add_points(np.array(d))
# set the viewer to 3D rendering mode with the first two rectangles in view
viewer.dims.ndisplay = 3
viewer.dims.set_point(axis=0, value=0)
viewer.camera.angles = (70, 30, 150)
viewer.camera.zoom = 2.5

if __name__ == '__main__':
    napari.run()



"""
Add surface 2D
==============

Display a 2D surface

.. tags:: visualization-basic
"""

import numpy as np

import napari

data = np.array([[0, 0], [0, 20], [10, 0], [10, 10]])
faces = np.array([[0, 1, 2], [1, 2, 3]])
values = np.linspace(0, 1, len(data))

# add the surface
viewer = napari.Viewer()
layer = viewer.add_surface((data , faces , values))

if __name__ == '__main__':
    napari.run()



"""
Layer Bounding Box (Border) Overlay
===================================

Display an image in napari and add a bounding box overlay (a border) at the edges of the layer.

The bounding box overlay is a visual representation of the extents of a layer.
This example demonstrates visualization of many bounding box overlay properties,
including line color, line thickness, point size, opacity.
In addition, this viewer shows how layer overlays interact when draw on top of each other (at the intersection of the grid).
Use `viewer.layers[0].bounding_box` to view all modifiable attributes of the overlay.

For an example showing how bounding box extents are visualized, see
:ref:`sphx_glr_gallery_layer_bounding_box.py`.

.. tags:: visualization-advanced
"""

from skimage import data

import napari

# Create a napari viewer with a 3D, 2 channel image, with the `add_image` convenience function.
viewer = napari.Viewer()
membrane, nuclei = viewer.add_image(data.cells3d(), channel_axis=1, name=['membrane', 'nuclei'])
viewer.grid.enabled = True

# Add a border overlay to each layer, and modify properties of the border overlay.
viewer.layers[0].bounding_box.visible = True
viewer.layers[0].bounding_box.line_color = 'cyan' # default: 'red'
viewer.layers[0].bounding_box.line_thickness = 5 # default: 1, maximum is GPU-dependent, see: https://vispy.org/api/vispy.scene.visuals.html#vispy.scene.visuals.Line
viewer.layers[0].bounding_box.point_size = 10 # default: 5
viewer.layers[0].bounding_box.point_color = 'yellow' # default: 'blue'

viewer.layers[1].bounding_box.visible = True
viewer.layers[1].bounding_box.line_color = 'orange'
viewer.layers[1].bounding_box.line_thickness = 2
viewer.layers[1].bounding_box.opacity = 0.8 # default: 1
# viewer_2d.layers[1].bounding_box.blending = 'additive' # default: 'translucent'

if __name__ == '__main__':
    napari.run()



"""
Add vectors
===========

This example generates an image of vectors
Vector data is an array of shape (N, 4)
Each vector position is defined by an (x, y, x-proj, y-proj) element where
* x and y are the center points
* x-proj and y-proj are the vector projections at each center

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data

import napari

# create the viewer and window
viewer = napari.Viewer()

layer = viewer.add_image(data.camera(), name='photographer')

# sample vector coord-like data
n = 200
pos = np.zeros((n, 2, 2), dtype=np.float32)
phi_space = np.linspace(0, 4 * np.pi, n)
radius_space = np.linspace(0, 100, n)

# assign x-y position
pos[:, 0, 0] = radius_space * np.cos(phi_space) + 300
pos[:, 0, 1] = radius_space * np.sin(phi_space) + 256

# assign x-y projection
pos[:, 1, 0] = 2 * radius_space * np.cos(phi_space)
pos[:, 1, 1] = 2 * radius_space * np.sin(phi_space)

# add the vectors
layer = viewer.add_vectors(pos, edge_width=3)

if __name__ == '__main__':
    napari.run()



"""
Grid mode
=========

Display layers in a grid using the `viewer.grid` API. When grid is enabled,
layers are automatically arranged - based on the stride, height, and width parameters -
in a grid of viewboxes linked to the main Camera and Dims. Viewer overlays such as
scale_bar can also be shown in a gridded manner.

.. tags:: visualization-basic
"""

from skimage import data

import napari

viewer = napari.Viewer()
layers = viewer.add_image(data.lily(), channel_axis=2)

viewer.grid.enabled = True
# a stride of 2 means that two consecutive layers are placed in each
# viewbox instead of just one
viewer.grid.stride = 2
# we leave width and height as -1, automatically resulting in a square grid

# setting the spacing to a value between 0 and 1 adds a padding between
# viewboxes relative to their sizes. Setting to a value greater than 1
# results in a padding of that exact number of pixels instead.
viewer.grid.spacing = 0.1

viewer.scale_bar.visible = True
viewer.scale_bar.box = True
# show scalebar in each grid instead of just once
viewer.scale_bar.gridded = True

if __name__ == '__main__':
    napari.run()



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://napari.org/developers/benchmarks.html

import numpy as np
import pandas as pd

from napari.layers.utils.text_manager import TextManager

from .utils import Skip


class TextManagerSuite:
    """Benchmarks for creating and modifying a text manager."""

    param_names = ['n', 'string']
    params = [
        [2**i for i in range(4, 18, 2)],
        [
            {'constant': 'test'},
            'string_property',
            'float_property',
            '{string_property}: {float_property:.2f}',
        ],
    ]

    skip_params = Skip(if_in_pr=lambda n, string: n > 2**6)

    def setup(self, n, string):
        np.random.seed(0)
        categories = ('cat', 'car')
        self.features = pd.DataFrame(
            {
                'string_property': pd.Series(
                    np.random.choice(categories, n),
                    dtype=pd.CategoricalDtype(categories),
                ),
                'float_property': np.random.rand(n),
            }
        )
        self.current_properties = self.features.iloc[[-1]].to_dict('list')
        self.manager = TextManager(string=string, features=self.features)
        self.indices_to_remove = list(range(0, n, 2))

    def time_create(self, n, string):
        TextManager(string=string, features=self.features)

    def time_refresh(self, n, string):
        self.manager.refresh_text(self.features)

    def time_add_iteratively(self, n, string):
        for _ in range(512):
            self.manager.add(self.current_properties, 1)

    def time_remove_as_batch(self, n, string):
        self.manager.remove(self.indices_to_remove)

    # `time_remove_as_batch` can only run once per instance;
    # otherwise it fails because the indices were already removed:
    #
    #   IndexError: index 32768 is out of bounds for axis 0 with size 32768
    #
    # Why? ASV will run the same function after setup several times in two
    # occasions: warmup and timing itself. We disable warmup and only
    # allow one execution per state with these method-specific options:
    time_remove_as_batch.number = 1
    time_remove_as_batch.warmup_time = 0
    # See https://asv.readthedocs.io/en/stable/benchmarks.html#timing-benchmarks
    # for more details


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



import os

import numpy as np

import napari

"""
This example generates many random shapes.

There currently is a bug in triangulation that requires this additional step of sanitizing the shape data: https://github.com/orgs/napari/projects/18/views/2
"""


# logging.getLogger().setLevel(0)


def generate_shapes(filename):
    # image_data = np.squeeze(cells3d()[:, 1, :, :])
    # delayed_image_data = DelayedArray(image_data, delay_s=1)

    # From https://github.com/napari/napari/blob/main/examples/nD_shapes.py
    # create one random polygon per "plane"

    shapes_per_slice = 1000

    all_shapes = None

    np.random.seed(0)
    for _ in range(shapes_per_slice):

        planes = np.tile(np.arange(128).reshape((128, 1, 1)), (1, 5, 1))
        corners = np.random.uniform(0, 128, size=(128, 5, 2))
        shapes = np.concatenate((planes, corners), axis=2)

        if all_shapes is not None:
            all_shapes = np.concatenate((all_shapes, shapes), axis=0)
        else:
            all_shapes = shapes

    print('all_shapes', all_shapes.shape)

    from vispy.geometry.polygon import PolygonData

    good_shapes = []

    for shape in all_shapes:
        # Use try/except to filter all bad shapes
        try:
            PolygonData(
                vertices=shape[:, 1:]
            ).triangulate()
        except AssertionError:
            pass
        else:
            good_shapes.append(shape)

    print(len(good_shapes))
    np.savez(filename, shapes=good_shapes)


test_filename = '/tmp/napari_example_shapes.npz'

# Create the example shapes if they do not exist
if not os.path.exists(test_filename):
    print(
        'Shapes file does not exist yet. Generating shapes. This may take a couple of minutes...'
    )
    generate_shapes(test_filename)

# Load the shapes
with np.load(test_filename) as data:
    shapes = data['shapes']

# Test shapes in viewer
viewer = napari.Viewer()
viewer.show()

shapes_layer = viewer.add_shapes(
    np.array(shapes),
    shape_type='polygon',
    name='sliced',
)

napari.run()



"""
Tracks 2D
=========

.. tags:: visualization-basic
"""

import numpy as np

import napari


def _circle(r, theta):
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def tracks_2d(num_tracks=10):
    """ create 2d+t track data """
    tracks = []

    for track_id in range(num_tracks):

        # space to store the track data and features
        track = np.zeros((100, 6), dtype=np.float32)

        # time
        timestamps = np.arange(track.shape[0])

        radius = 20 + 30 * np.random.random()
        theta = timestamps * 0.1 + np.random.random() * np.pi
        x, y = _circle(radius, theta)

        track[:, 0] = track_id
        track[:, 1] = timestamps
        track[:, 2] = 50.0 + y
        track[:, 3] = 50.0 + x
        track[:, 4] = theta
        track[:, 5] = radius

        tracks.append(track)

    tracks = np.concatenate(tracks, axis=0)
    data = tracks[:, :4]  # just the coordinate data

    features = {
        'time': tracks[:, 1],
        'theta': tracks[:, 4],
        'radius': tracks[:, 5],
    }

    graph = {}
    return data, features, graph


tracks, features, graph = tracks_2d(num_tracks=10)
vertices = tracks[:, 1:]

viewer = napari.Viewer()
viewer.add_points(vertices, size=1, name='points', opacity=0.3)
viewer.add_tracks(tracks, features=features, name='tracks')

if __name__ == '__main__':
    napari.run()



"""
Get current viewer
==================

Get a reference to the current napari viewer.

Whilst this example is contrived, it can be useful to get a reference to the
viewer when the viewer is out of scope.

.. tags:: gui
"""
import numpy as np

import napari

# create viewer
viewer = napari.Viewer()

# lose reference to viewer
viewer = 'oops no viewer here'

# get that reference again
viewer = napari.current_viewer()

# work with the viewer
x = np.arange(256)
y = np.arange(256).reshape((256, 1))
# from: https://botsin.space/@bitartbot/113553754823363986
image = (-(~((y - x) ^ (y + x)))) % 11
layer = viewer.add_image(image)
layer.contrast_limits = (8.5, 10)

if __name__ == '__main__':
    napari.run()



"""
3D Kymographs
=============

This example demonstrates that the volume rendering capabilities of napari
can also be used to render 2d timelapse acquisitions as kymographs.

.. tags:: experimental
"""
from itertools import product

import numpy as np

try:
    from omero.gateway import BlitzGateway
except ModuleNotFoundError:
    print('Could not import BlitzGateway which is')
    print('required to download the sample datasets.')
    print('Please install omero-py:')
    print('https://pypi.org/project/omero-py/')
    exit(-1)
try:
    from tqdm import tqdm
except ModuleNotFoundError:
    print('Could not import tqdm which is')
    print('required to show progress when downloading the sample datasets.')
    print('Please install tqdm:')
    print('https://pypi.org/project/tqdm/')
    exit(-1)

import napari


def IDR_fetch_image(image_id: int, progressbar: bool = True) -> np.ndarray:
    """
    Download the image with id image_id from the IDR

    Will fetch all image planes corresponding to separate
    timepoints/channels/z-slices and return a numpy
    array with dimension order (t,z,y,x,c)

    Displaying download progress can be disabled by passing
    False to progressbar.
    """

    conn = BlitzGateway(
        host='ws://idr.openmicroscopy.org/omero-ws',
        username='public',
        passwd='public',
        secure=True,
    )
    conn.connect()
    conn.c.enableKeepAlive(60)

    idr_img = conn.getObject('Image', image_id)
    idr_pixels = idr_img.getPrimaryPixels()

    _ = idr_img
    nt, nz, ny, nx, nc = (
        _.getSizeT(),
        _.getSizeZ(),
        _.getSizeY(),
        _.getSizeX(),
        _.getSizeC(),
    )

    plane_indices = list(product(range(nz), range(nc), range(nt)))
    idr_plane_iterator = idr_pixels.getPlanes(plane_indices)

    if progressbar:
        idr_plane_iterator = tqdm(idr_plane_iterator, total=len(plane_indices))

    _tmp = np.asarray(list(idr_plane_iterator))
    _tmp = _tmp.reshape((nz, nc, nt, ny, nx))
    # the following line reorders the axes (no summing, despite the name)
    return np.einsum('jmikl', _tmp)


description = """
3D-Kymographs in Napari
=======================

About
=====
This example demonstrates that the volume rendering capabilities of napari
can also be used to render 2d timelapse acquisitions as kymographs.
Kymographs, also called space-time images, are a powerful tool to visualize
the dynamics of processes.
The most common way to visualize kymographs is to pick a single line through
a 2D image and visualize the time domain along a second axes.
Napari is not limited to 2D visualization an by harnessing its volume
volume rendering capabilities, we can create a 3D kymograph,
a powerful visualization that provides an overview of the complete
spatial and temporal data from a single view.

Using napari's grid mode we can juxtapose multiple such 3D kymographs to
highlight the differences in cell dynamics under different siRNA treatments.

The selected samples are from the Mitocheck screen and demonstrate siRNA
knockdowns of several genes.
The date is timelapse fluorescence microscopy of HeLa cells, with GFP-
tagged histone revealing the chromosomes.

In the juxtaposed kymographs the reduced branching for the mitotitic
phenotypes caused by INCENP, AURKB and KIF11 knockdown compared to
TMPRSS11A knockdown is immediately obvious.

Data Source
===========
The samples to demonstrate this is downloaded from IDR:
https://idr.openmicroscopy.org/webclient/?show=screen-1302

Reference
=========
The data comes from the Mitocheck screen:

Phenotypic profiling of the human genome by time-lapse microscopy reveals cell
division genes.

Neumann B, Walter T, Hériché JK, Bulkescher J, Erfle H, Conrad C, Rogers P,
Poser I, Held M, Liebel U, Cetin C, Sieckmann F, Pau G, Kabbe R, Wünsche A,
Satagopam V, Schmitz MH, Chapuis C, Gerlich DW, Schneider R, Eils R, Huber W,
Peters JM, Hyman AA, Durbin R, Pepperkok R, Ellenberg J.
Nature. 2010 Apr 1;464(7289):721-7.
doi: 10.1038/nature08869.

Acknowledgements
================
Beate Neumann (EMBL) for helpful advice on mitotic phenotypes.

"""

print(description)

samples = (
    {'IDRid': 2864587, 'description': 'AURKB knockdown', 'vol': None},
    {'IDRid': 2862565, 'description': 'KIF11 knockdown', 'vol': None},
    {'IDRid': 2867896, 'description': 'INCENP knockdown', 'vol': None},
    {'IDRid': 1486532, 'description': 'TMPRSS11A knockdown', 'vol': None},
)

print('-------------------------------------------------------')
print('Sample datasets will require ~490 MB download from IDR.')
answer = input("Press Enter to proceed, 'n' to cancel: ")
if answer.lower().startswith('n'):
    print('User cancelled download. Exiting.')
    exit(0)
print('-------------------------------------------------------')
for s in samples:
    print(f"Downloading sample {s['IDRid']}.")
    print(f"Description: {s['description']}")
    s['vol'] = np.squeeze(IDR_fetch_image(s['IDRid']))

v = napari.Viewer(ndisplay=3)
scale = (5, 1, 1)  # "stretch" time domain
for s in samples:
    v.add_image(
        s['vol'], name=s['description'], scale=scale, blending='opaque'
    )

v.grid.enabled = True  # show the volumes in grid mode
v.axes.visible = True  # magenta error shows time direction

# set an oblique view angle onto the kymograph grid
v.camera.center = (440, 880, 1490)
v.camera.angles = (-20, 23, -50)
v.camera.zoom = 0.17

napari.run()



from napari._qt.layer_controls.widgets._shapes.qt_edge_color import (
    QtEdgeColorControl,
)
from napari._qt.layer_controls.widgets._shapes.qt_edge_width_slider import (
    QtEdgeWidthSliderControl,
)

__all__ = ['QtEdgeColorControl', 'QtEdgeWidthSliderControl']



from qtpy.QtWidgets import QWidget
from superqt import QEnumComboBox

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Labels
from napari.layers.labels._labels_constants import (
    IsoCategoricalGradientMode,
    LabelsRendering,
)
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtLabelRenderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer attribute for
    the method to render the labels and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels Labels layer.

    Attributes
    ----------
    iso_gradient_combobox : superqt.QEnumComboBox
        Combobox to control gradient method when isosurface rendering is selected.
    iso_gradient_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the chooser widget of the gradient to use when labels are using isosurface rendering.
    render_combobox : superqt.QEnumComboBox
        Combobox to control current label render method.
    render_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the way labels should be rendered chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.rendering.connect(self._on_rendering_change)
        self._layer.events.iso_gradient_mode.connect(
            self._on_iso_gradient_mode_change
        )

        # Setup widgets
        render_combobox = QEnumComboBox(enum_class=LabelsRendering)
        render_combobox.setCurrentEnum(LabelsRendering(self._layer.rendering))
        self.render_combobox = render_combobox
        connect_setattr(
            render_combobox.currentEnumChanged, self._layer, 'rendering'
        )
        self.render_combobox_label = QtWrappedLabel(trans._('rendering:'))

        iso_gradient_combobox = QEnumComboBox(
            enum_class=IsoCategoricalGradientMode
        )
        iso_gradient_combobox.setCurrentEnum(
            IsoCategoricalGradientMode(self._layer.iso_gradient_mode)
        )
        connect_setattr(
            iso_gradient_combobox.currentEnumChanged,
            self._layer,
            'iso_gradient_mode',
        )
        iso_gradient_combobox.setEnabled(
            self._layer.rendering == LabelsRendering.ISO_CATEGORICAL
        )
        self.iso_gradient_combobox = iso_gradient_combobox
        self.iso_gradient_combobox_label = QtWrappedLabel(
            trans._('gradient\nmode:')
        )

    def _on_rendering_change(self):
        """Receive layer model rendering change event and update dropdown menu."""
        rendering_mode = LabelsRendering(self._layer.rendering)

        with qt_signals_blocked(self.render_combobox):
            self.render_combobox.setCurrentEnum(rendering_mode)

        with qt_signals_blocked(self.iso_gradient_combobox):
            self.iso_gradient_combobox.setEnabled(
                rendering_mode == LabelsRendering.ISO_CATEGORICAL
            )

    def _on_iso_gradient_mode_change(self):
        """Receive layer model iso_gradient_mode change event and update dropdown menu."""
        with qt_signals_blocked(self.iso_gradient_combobox):
            self.iso_gradient_combobox.setCurrentEnum(
                IsoCategoricalGradientMode(self._layer.iso_gradient_mode)
            )

    def _on_display_change_hide(self):
        self.render_combobox.hide()
        self.render_combobox_label.hide()
        self.iso_gradient_combobox.hide()
        self.iso_gradient_combobox_label.hide()

    def _on_display_change_show(self):
        self.render_combobox.show()
        self.render_combobox_label.show()
        self.iso_gradient_combobox.show()
        self.iso_gradient_combobox_label.show()

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.render_combobox_label, self.render_combobox),
            (self.iso_gradient_combobox_label, self.iso_gradient_combobox),
        ]



"""
Features table widget
======================

Visualize and interact with the `features` of a layer via the builtin
Features table widget.
The widget can be used to navigate, edit, select, copy/paste, and save the features
of a compatible layer (such as Points, Labels or Shapes).
Selection is synchronized between the table widget and the data. When selecting rows in
the table widget, the corresponding data will be selected on the layer. When selecting
data on the layer, the corresponding row in the table widget will be selected.
on the layer and viceversa.

If multiple layers are selected, their features will be joined and displayed together in the table.
The layer name column indicates the origin layer of each feature row and a "shared columns" toggle
allows to switch between showing only the columns shared by all selected layers or all columns.

.. tags:: gui, features-table
"""

import numpy as np
import pandas as pd

import napari

viewer = napari.Viewer(ndisplay=3)

features = pd.DataFrame({
    'a': np.random.rand(10),
    'b': ['stuff'] * 10,
    'c': np.random.choice([0, 1], 10).astype(bool),
    'd': pd.Series(np.random.choice(['x', 'y', 'z'], 10), dtype='category')
})
features.loc[3, 'a'] = np.nan
features.loc[2, 'b'] = 'asd'

features2 = pd.DataFrame({
    'a': np.random.rand(10) + 2,
    'f': np.random.randint(0, 100, 10),
})
viewer.add_image(np.random.rand(10, 10))

points = np.random.rand(10, 3) * 10
faces = np.array([[0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, 5],
                   [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9]])
surface_layer = viewer.add_surface((points, faces, features2['a'].values), features=features2, colormap='plasma')
points_layer = viewer.add_points(points, size=1, features=features, name='Points')

viewer.layers.selection.add(points_layer)
viewer.layers.selection.add(surface_layer)
viewer.window.add_plugin_dock_widget('napari', 'Features table widget')

if __name__ == '__main__':
    napari.run()



"""
Xarray example
==============

Displays an xarray

.. tags:: visualization-nD, xarray
"""

try:
    import xarray as xr
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        """This example uses a xarray but xarray is not
    installed. To install try 'pip install xarray'."""
    ) from None

import numpy as np

import napari

data = np.random.random((20, 40, 50))
xdata = xr.DataArray(data, dims=['z', 'y', 'x'])

# create an empty viewer
viewer = napari.Viewer()

# add the xarray
layer = viewer.add_image(xdata, name='xarray')

if __name__ == '__main__':
    napari.run()



"""
Add image transformed
=====================

Display one image and transform it using the :func:`add_image` API.

.. tags:: visualization-basic
"""

from skimage import data

import napari

# create the viewer with an image and transform (rotate) it
viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True, rotate=45)

if __name__ == '__main__':
    napari.run()



"""
Edit pull request description to remove HTML comments

We might want to remove section with markdown task lists that are completely empty
"""

import re
import sys
from os import environ

import requests

REPO = 'napari/napari'


def remove_html_comments(text):
    # Regular expression to remove HTML comments
    # [^\S\r\n] is whitespace but not new line
    html_comment_pattern = r'[^\S\r\n]*<!--(.*?)-->[^\S\r\n]*\s*'
    return re.sub(html_comment_pattern, '\n', text, flags=re.DOTALL)


def edit_pull_request_description(repo, pull_request_number, access_token):
    # GitHub API base URL
    base_url = 'https://api.github.com'

    # Prepare the headers with the access token
    headers = {'Authorization': f'token {access_token}'}

    # Get the current pull request description
    pr_url = f'{base_url}/repos/{repo}/pulls/{pull_request_number}'
    response = requests.get(pr_url, headers=headers)
    response.raise_for_status()
    response_json = response.json()
    current_description = response_json['body']

    # Remove HTML comments from the description
    edited_description = remove_html_comments(current_description)

    if edited_description == current_description:
        print('No HTML comments found in the pull request description')
        return

    # Update the pull request description
    update_pr_url = f'{base_url}/repos/{repo}/pulls/{pull_request_number}'
    payload = {'body': edited_description}
    response = requests.patch(update_pr_url, json=payload, headers=headers)
    response.raise_for_status()

    if response.status_code == 200:
        print(
            f'Pull request #{pull_request_number} description has been updated successfully!'
        )
    else:
        print(
            f'Failed to update pull request description. Status code: {response.status_code}'
        )


if __name__ == '__main__':
    print('Will inspect PR description to remove html comments.')

    # note that the env between pull_request and pull_request_target are different
    # and the github documentation is incorrect (or at least misleading)
    # and likely varies between pull request intra-repository and inter-repository
    # thus we log many things to try to understand what is going on in case of failure.
    # among other:
    # - github.event.repository.name is not the full slug, but just the name
    # - github.event.repository.org is empty if the repo is a normal user.

    repository_url = environ.get('GH_REPO_URL')
    print(f'Current repository is {repository_url}')
    repository_parts = repository_url.split('/')[-2:]

    slug = '/'.join(repository_parts)
    print(f'Current slug is {slug}')
    if slug != REPO:
        print('Not on main repo, aborting with success')
        sys.exit(0)

    # get current PR number from github actions
    number = environ.get('GH_PR_NUMBER')
    print(f'Current PR number is {number}')

    access_token = environ.get('GH_TOKEN')
    if access_token is None:
        print('No access token found in the environment variables')
        # we still don't want fail status
        sys.exit(0)
    edit_pull_request_description(slug, number, access_token)



import argparse

import numpy as np

import napari

"""
Stress the points layer by generating a large number of points.
"""

parser = argparse.ArgumentParser()
parser.add_argument(
    'n', type=int, nargs='?', default=10_000_000, help='(default: %(default)s)'
)
args = parser.parse_args()

np.random.seed(0)
n = args.n
data = 1000 * np.random.rand(n, 3)
viewer = napari.Viewer()
layer = viewer.add_points(data)
if __name__ == '__main__':
    napari.run()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

import numpy as np
import pytest
from qtpy.QtCore import QByteArray, QObject, Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QColorDialog, QMainWindow

from napari._qt.utils import (
    QBYTE_FLAG,
    add_flash_animation,
    get_color,
    is_qbyte,
    qbytearray_to_str,
    qt_might_be_rich_text,
    qt_signals_blocked,
    str_to_qbytearray,
)
from napari.utils._proxies import PublicOnlyProxy

if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot


class Emitter(QObject):
    test_signal = Signal()

    def go(self):
        self.test_signal.emit()


def test_signal_blocker(qtbot):
    """make sure context manager signal blocker works"""

    obj = Emitter()

    # make sure signal works
    with qtbot.waitSignal(obj.test_signal):
        obj.go()

    # make sure blocker works
    with (
        qt_signals_blocked(obj),
        qtbot.assert_not_emitted(obj.test_signal, wait=500),
    ):
        obj.go()
    obj.deleteLater()


def test_is_qbyte_valid():
    assert is_qbyte(QBYTE_FLAG)
    assert is_qbyte(
        '!QBYTE_AAAA/wAAAAD9AAAAAgAAAAAAAAECAAACePwCAAAAAvsAAAAcAGwAYQB5AGUAcgAgAGMAbwBuAHQAcgBvAGwAcwEAAAAAAAABFwAAARcAAAEX+wAAABQAbABhAHkAZQByACAAbABpAHMAdAEAAAEXAAABYQAAALcA////AAAAAwAAAAAAAAAA/AEAAAAB+wAAAA4AYwBvAG4AcwBvAGwAZQAAAAAA/////wAAADIA////AAADPAAAAngAAAAEAAAABAAAAAgAAAAI/AAAAAA='
    )


def test_str_to_qbytearray_valid():
    assert isinstance(
        str_to_qbytearray(
            '!QBYTE_AAAA/wAAAAD9AAAAAgAAAAAAAAECAAACePwCAAAAAvsAAAAcAGwAYQB5AGUAcgAgAGMAbwBuAHQAcgBvAGwAcwEAAAAAAAABFwAAARcAAAEX+wAAABQAbABhAHkAZQByACAAbABpAHMAdAEAAAEXAAABYQAAALcA////AAAAAwAAAAAAAAAA/AEAAAAB+wAAAA4AYwBvAG4AcwBvAGwAZQAAAAAA/////wAAADIA////AAADPAAAAngAAAAEAAAABAAAAAgAAAAI/AAAAAA='
        ),
        QByteArray,
    )


def test_str_to_qbytearray_invalid():
    with pytest.raises(ValueError, match='Invalid QByte string'):
        str_to_qbytearray('')

    with pytest.raises(ValueError, match='Invalid QByte string'):
        str_to_qbytearray('FOOBAR')

    with pytest.raises(ValueError, match='Invalid QByte string'):
        str_to_qbytearray(
            '_AAAA/wAAAAD9AAAAAgAAAAAAAAECAAACePwCAAAAAvsAAAAcAGwAYQB5AGUAcgAgAGMAbwBuAHQAcgBvAGwAcwEAAAAAAAABFwAAARcAAAEX+wAAABQAbABhAHkAZQByACAAbABpAHMAdAEAAAEXAAABYQAAALcA////AAAAAwAAAAAAAAAA/AEAAAAB+wAAAA4AYwBvAG4AcwBvAGwAZQAAAAAA/////wAAADIA////AAADPAAAAngAAAAEAAAABAAAAAgAAAAI/AAAAAA='
        )


def test_qbytearray_to_str(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)

    qbyte = widget.saveState()
    qbyte_string = qbytearray_to_str(qbyte)
    assert is_qbyte(qbyte_string)


def test_qbytearray_to_str_and_back(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)

    qbyte = widget.saveState()
    assert str_to_qbytearray(qbytearray_to_str(qbyte)) == qbyte


def test_add_flash_animation(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)
    assert widget.graphicsEffect() is None
    add_flash_animation(widget)
    assert widget.graphicsEffect() is not None
    assert hasattr(widget, '_flash_animation')
    qtbot.wait(350)
    assert widget.graphicsEffect() is None
    assert not hasattr(widget, '_flash_animation')


@pytest.mark.usefixtures('qapp')
def test_qt_might_be_rich_text():
    assert qt_might_be_rich_text('<b>rich text</b>')
    assert not qt_might_be_rich_text('plain text')


@pytest.mark.usefixtures('qapp')
def test_thread_proxy_guard(monkeypatch, single_threaded_executor):
    class X:
        a = 1

    monkeypatch.setenv('NAPARI_ENSURE_PLUGIN_MAIN_THREAD', 'True')

    x = X()
    x_proxy = PublicOnlyProxy(x)

    f = single_threaded_executor.submit(x.__setattr__, 'a', 2)
    f.result()
    assert x.a == 2

    f = single_threaded_executor.submit(x_proxy.__setattr__, 'a', 3)
    with pytest.raises(RuntimeError):
        f.result()
    assert x.a == 2


def _assert_eq(p1: Any, p2: Any, text: str = '') -> None:
    if isinstance(p2, np.ndarray):
        np.testing.assert_array_equal(p1, p2, err_msg=text)
    else:
        assert p1 == p2, text


DEFAULT_COLOR_HEX = '#ffffff'
DEFAULT_COLOR_ARRAY = np.asarray([1, 1, 1])
DEFAULT_COLOR_QCOLOR = QColor(255, 255, 255)


@pytest.mark.parametrize(
    ('color', 'mode', 'expected'),
    [
        (None, 'hex', DEFAULT_COLOR_HEX),
        ('#FF00FF', 'hex', '#ff00ff'),
        (None, 'array', DEFAULT_COLOR_ARRAY),
        (np.asarray([255, 0, 255]), 'array', np.asarray([1, 0, 1])),
        (None, 'qcolor', DEFAULT_COLOR_QCOLOR),
    ],
)
def test_get_color(
    qtbot: QtBot,
    color: str | np.ndarray | None,
    mode: Literal['hex', 'array', 'qcolor'],
    expected: str | np.ndarray | QColor,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test the get_color utility function."""

    def _mock_exec_(self):
        """Mock exec_ method to always return Accepted."""
        qtbot.addWidget(self)
        return QColorDialog.DialogCode.Accepted

    widget = QMainWindow()
    qtbot.addWidget(widget)

    monkeypatch.setattr(QColorDialog, 'exec_', _mock_exec_)

    color_ = get_color(color, mode)
    _assert_eq(color_, expected, f'Expected color to be {expected}')


@pytest.mark.parametrize(
    ('color', 'mode'),
    [
        (None, 'hex'),
        ('#FF00FF', 'hex'),
        (None, 'array'),
        (np.asarray([255, 0, 255]), 'array'),
        (None, 'qcolor'),
    ],
)
def test_get_color_reject(
    qtbot: QtBot,
    color: str | np.ndarray | None,
    mode: Literal['hex', 'array', 'qcolor'],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test the get_color utility function."""

    def _mock_exec_(self):
        """Mock exec_ method to always return Accepted."""
        qtbot.addWidget(self)
        return QColorDialog.DialogCode.Rejected

    widget = QMainWindow()
    qtbot.addWidget(widget)

    monkeypatch.setattr(QColorDialog, 'exec_', _mock_exec_)

    color_ = get_color(color, mode)
    assert color_ is None, 'Expected color to be None when dialog is rejected'



"""
Add points 3D
=============

Display a labels layer above of an image layer using the add_labels and
add_image APIs, then add points in 3D

.. tags:: visualization-nD
"""

from scipy import ndimage as ndi
from skimage import data

import napari

blobs = data.binary_blobs(
        length=128, volume_fraction=0.1, n_dim=3
        )[::2].astype(float)
labeled = ndi.label(blobs)[0]

viewer = napari.Viewer(ndisplay=3)
viewer.add_image(blobs, name='blobs', scale=(2, 1, 1))
viewer.add_labels(labeled, name='blob ID', scale=(2, 1, 1))
pts = viewer.add_points()

viewer.camera.angles = (0, -65, 85)
pts.mode = 'add'

if __name__ == '__main__':
    napari.run()
